#ifndef _GOSTC_ASYM_H_
#define _GOSTC_ASYM_H_

#include "gostc_common.h"
#include "gostc_asym_opers.h"

#ifdef __cplusplus
extern "C" {
#endif

void*	gostc_asym_LoadParams(int n);
/* n - 1-TestGost, 2-CproA(XchA), 3-CproB, 4-CproC(XchB), 5-512bit-TestParamSet,
       6-512bit-ParamSetA, 7-512bit-ParamSetB */
/*ret NULL, if error, or pointer, if OK*/

/* int	gostc_asym_ReloadParams(void* params, int inNum); */
/* ret: 0-err, 1-ok */
/* Commented out, because not used and erroneous */


/********************************************************************************/
/*for next funcs:	params - CRYPTO CONTEXT! Must not be NULL!		*/
/********************************************************************************/
void	gostc_asym_FreeParams(void *params);

int	gostc_asym_MakeSign(void* params, uint8_t* rnd32_64b, void *h, void *keyD, void *sign);
/* Computing sign. ret: 0-err, 1-ok, 2-more rnd needed, ##
 * 	
b - bit mode :	rnd32_64b-random	h - hash, 	keyD - secret key, 	sign - out for sign.
size for 256b:	32 bytes,		32 bytes,	32 bytes,		64 bytes
size for 512b: 	64 bytes,		64 bytes,	64 bytes,		128 bytes 
*/

int	gostc_asym_CheckSign(void* params, uint8_t* keyQ, uint8_t* hash, uint8_t* sign);
/* Checking sign: 0-wrong sign, 1-ok, 2-err 	##*/
/* 	
b - bit mode :	keyQ-public key,	h - hash, 	sign - checking sign.
size for 256b:	64 bytes,		32 bytes,	64 bytes,		
size for 512b: 	128 bytes,		64 bytes,	128 bytes,		
*/
int	gostc_asym_GenSecretKey(void* params, uint8_t* rnd32_64b, void* keyD);
/* ret: 0-err, 1-ok, 2-more rnd needed ##
 * rnd32_64b - random:		32 bytes for 256 bit mode, 64 bytes for 512 bit mode */
/*	keyD - out for key: 	32 bytes for 256 bit mode, 64 bytes for 512 bit mode */

int	gostc_asym_GenPublicKey(void* params, void* keyD, void* keyQ);
/*	Generating public key*/
/* ret: 0-err, 1-ok ##*/
/*	keyD - input secret key: 32 bytes for 256 bit mode, 64 bytes for 512 bit mode 	*/
/*	keyQ - output public key:64 bytes for 256 bit mode, 128 bytes for 512 bit mode 	*/



/*Computing VKO-key as a Diffie-Hellman's key	*/
/*UKM - see below */
/*	keyD - input secret key: 32 bytes for 256 bit mode, 64 bytes for 512 bit mode 	*/
/*	keyQ - input public key: 64 bytes for 256 bit mode, 128 bytes for 512 bit mode 	*/
/*	keyDFH - out for VKO:	 64 bytes for 256 bit mode, 128 bytes for 512 bit mode 	*/
int	gostc_asym_GenDFHKey(void* params, void* keyD, void* keyQ, void * UKM, void* keyDFH);
/* ret: 0-err, 1-ok ##*/
/* UKM - 8 bytes for 256 bit mode, 16 bytes for 512 bit mode */
/* Old interface preserved for compatibility */

int	gostc_asym_GenDFHKey2(void* params, void* keyD, void* keyQ, 
	void* UKM, size_t UKMlen, void* keyDFH);
/* New interface: UKMlen specified and may be 8 or 16 for any paramset */
/* ret: 0-err, 1-ok ## */


int	gostc_asym_TestLib(void);
/* ret: 0-lib fail, 1-lib ok */



#ifdef __cplusplus
}
#endif



#endif

