#ifndef _GOSTC_ASYM_OPERS_H_
#define _GOSTC_ASYM_OPERS_H_


#include "gostc_asym_elems.h"

#ifdef __cplusplus
extern "C" {
#endif


/*##############################################################################*/
/*   Надстройки над элементарными математическими операциями                  	*/
/*	здесь и далее params - структуры 256 и 512-битного контекста.		*/
/*	О разрядности чисел см. gostc_asym_elems.h				*/
/*##############################################################################*/
/*   Возврат процедур: 0 - ОК, <0 - ошибки						*/
/*****Процедуры вычисления выражений S = A*B с приведением по модулю P целочисленным делением*************/
/*gostc_FindEQABModP поддерживает работу с 8-разрядными числами, gostc_Find16ModP с 16-разрядными	*/
int gostc_FindEQABModP(uint32_t *A,uint32_t *B,const uint32_t *p,uint32_t *S);
int gostc_Find16ModP(uint32_t *A,uint32_t *B,const uint32_t *p,uint32_t *S);

/*****Процедуры вычисления выражений S = A*B с приведением по модулю P в представлении Монтгомери****************/
/*gostc_FindEQABModPinM поддерживает работу с 8-разрядными числами, gostc_Find16ModPinM с 16-разрядными		*/
int gostc_FindEQABModPinM(const gostc_CR_SYSTEM *params,const uint32_t *A,const uint32_t *B,const uint32_t *p,uint32_t *S);
int gostc_Find16ModPinM(const gostc_CR_SYSTEM64 *params,const uint32_t *A,const uint32_t *B, const uint32_t *p,uint32_t *S);

/*****Процедуры сложения Rez = A + B с приведением по модулю P *************	*/
/*gostc_AplusBModP_C - 8-разрядные числа, gostc_AplusBModP64 - 16-разрядные	*/
int gostc_AplusBModP_C(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez);
int gostc_AplusBModP64(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez);


/******Процедуры вычитания Rez = A - B с приведением по модулю P   ********		*/
/*gostc_AMinusBModP_C - 8-разрядные числа, gostc_AMinusBModP64 - 16-разрядные		*/
/*gostc_SimpleSub - надстройка над gostc_AMinusBModP_C: A = (A - B)mod p	 	*/
int gostc_SimpleSub(uint32_t *A,uint32_t *B,const uint32_t *P);
int gostc_AMinusBModP_C(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez);
int gostc_AMinusBModP64(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez);

/******Процедуры вычисления Rez = (A - B)/2 с приведением по модулю P   ********	*/
/*halfP - число P/2 в представлении Монтгомери						*/
/*gostc_HALFAMinusBModPinM - 8-разрядные числа, gostc_HALFAMinusB64 - 16-разрядные	*/

int gostc_HALFAMinusBModPinM(uint32_t *A,uint32_t *B,const uint32_t *P,const uint32_t *halfP,uint32_t *Rez);
int gostc_HALFAMinusB64(uint32_t *A,uint32_t *B,const uint32_t *P, const uint32_t *halfP,uint32_t *Rez);

/****Процедура поиска числа Rez = 1/a по модулю P (Бинарный алгоритм Евклида)*		*/
/*gostc_BinaryEuclid_C - 8-разрядные числа, gostc_BinaryEuclid64 - 16-разрядные числа	*/
int gostc_BinaryEuclid_C(uint32_t *a, const uint32_t *p, uint32_t *rez);
int gostc_BinaryEuclid64(uint32_t *a, const uint32_t *p, uint32_t *rez);



/*Сложение двух точек (для 256 и 512-битных наборов) 
в проективных координатах pa и pb(возврат: 0 - ОК, <0 - ошибка). Результат пишется в pRez */
int gostc_AddingInProjectiveWInM(gostc_pONLY_POINT pa, gostc_pONLY_POINT pb,
	gostc_pONLY_POINT pRez, const gostc_CR_SYSTEM* params);
int gostc_AddInProj64M(gostc_pPOINT64 pa,gostc_pPOINT64 pb,gostc_pPOINT64 pRez,const gostc_CR_SYSTEM64* params);

/*Удвоение точки pa (для 256 и 512-битных наборов) в проективных координатах: 
Если pRez == NULL, то результат сохраняется в pa
возврат: 0 - ОК, <0 - ошибка*/
int gostc_DoubleInProjectiveWM(const gostc_CR_SYSTEM* params, gostc_pONLY_POINT pa, gostc_pONLY_POINT pRez);
int gostc_DoubleInProj64M(const gostc_CR_SYSTEM64* params, gostc_pPOINT64 pa, gostc_pPOINT64 pRez);

/*Поиск кратной точки C = kP (для 256 и 512-битных наборов), mas - представление числа k 
Если noconvert=1 - не выводить результат из представления Монтгомери в обычное представление
Вычисленные координаты точки C сохраняются в p1 
Возврат: 0 - ОК, <0 - ошибка 								*/
/*InM		- с использованием предвычисленных кратных точек, 8-разрядные числа */
/*64		- 16-разрядные числа */
int gostc_FindKrPointInM(const gostc_CR_SYSTEM* params, uint8_t *mas,int noconvert, gostc_POINT_ECP *p1);
int gostc_FindKrPoint64(const gostc_CR_SYSTEM64* params, uint8_t *mas,int noconvert, gostc_POINT_ECP64 *p1);

/*Нахождение суммы кратных точек(для 256 и 512-битных наборов) 
 k1P + k2Q, QM0 - представление точки Q, mas1 и mas2 - представления чисел k1 и k2
Вычисленная x-координата cуммы сохраняются в x_result 
Возврат: 0 - ОК, <0 - ошибка 	*/
int gostc_FindSumOfKrPoints(const gostc_CR_SYSTEM* params, gostc_pONLY_POINT QM0, uint32_t *mas1, uint32_t *mas2, uint32_t *x_result);
int gostc_FindSumKrPoints64(const gostc_CR_SYSTEM64* params, gostc_pPOINT64 QM0, uint8_t *mas1, uint8_t *mas2, uint32_t *x_result);

/*Нахождение кратной точки (для 256 и 512-битных наборов) по алгоритму Диффи-Хэллмана: mas*pbase*/
/*Результат сохраняется в prez					*/
/*Возврат: 0 - ОК, <0 - ошибка 	*/
int gostc_FindKrPointDiffiHellman(const gostc_CR_SYSTEM *params, uint8_t *mas,gostc_pONLY_POINT pbase, gostc_POINT_ECP *prez);
int gostc_FindKrPointDiffiHellman64(const gostc_CR_SYSTEM64 *params, uint8_t *mas,gostc_pPOINT64 pbase, gostc_POINT_ECP64 *prez);

/*##############################################################################*/
/*Далее для процедур более высокого уровня возврат: 1 - ОК, != 1 - ошибка	*/
/*##############################################################################*/
/*Процедуры вычисления ключей для ВКО (32 байта и 64 байта). Возврат: 0 - ошибка, 1 - ОК
par - криптонабор, keyD - "свой" закрытый ключ, ukm16 - UKM, keyQ - "чужой" открытый, результат сохраняется в keyDFH*/
int gostc_GenDFHKey32(const gostc_CR_SYSTEM *par,void* keyD,void * ukm16, void* keyQ,void* keyDFH);
int gostc_GenDFHKey64(const gostc_CR_SYSTEM64* par,void* keyD,void *ukm16,void* keyQ,void* keyDFH);


/*Процедуры наложения/снятия маски mask на ключ key, n - номер криптонабора 4-6(!!!, не 5-7, как в Init...) - 512 бит
 Возврат: 0 - ошибка, 1 - ОК							*/
int gostc_DoMask64(void *key, void* mask,uint32_t n);
int gostc_DoUnmask64(void *key, void* mask,uint32_t n);

#ifdef __cplusplus
}
#endif

#endif

