#include "gostc_mem.h"
#include "gostc_asym.h"
#include "gostc_asym_params.h" 


/*#define GOSTC_ASYM_TEST_SPEED */

#ifdef GOSTC_ASYM_TEST_SPEED
#include <stdlib.h>
#include <stdio.h>  /* GOSTC_ASYM_TEST_SPEED */
#include <sys/time.h>
#endif

/*
UCHAR testUZ[64] = 
{0xe4,0x75,0x46,0x1d,0xba,0xd8,0xbc,0xfb,
0x49,0xa1,0xa7,0xd4,0xc2,0x1d,0x01,0x01,
0x6d,0x0a,0x75,0x53,0xd8,0x83,0x2f,0x7f,
0xf0,0x94,0x1d,0xa5,0xae,0xf2,0xd8,0x49,
0x26,0xee,0x34,0x90,0x3b,0x4f,0x6a,0x2a,
0x81,0x6c,0x89,0x3e,0x1c,0xc7,0x5e,0xe7,
0x07,0xb6,0x90,0x66,0x7f,0x20,0xc3,0xb8,
0x55,0x59,0xfb,0x82,0x93,0x3b,0xe2,0xcc};

UCHAR CSPUZ[64] = 
{0x5A,0x47,0x77,0x1D,0xF4,0xAF,0x66,0x3E,0x45,0x7C,0x24,0xA4,0x06,0xCE,0x4B,0x91,
 0x28,0x09,0xD9,0x57,0xD1,0xF4,0x9C,0xB0,0xB3,0x21,0xF2,0x45,0x97,0x80,0x0A,0xFA,
 0x1D,0xE3,0xA1,0x83,0x7C,0x1B,0x18,0x6C,0x6E,0x65,0x50,0x78,0x30,0x52,0xBE,0xEF,
 0xC9,0xD6,0x8F,0xD6,0xE2,0xBA,0xED,0x02,0xAB,0x98,0xC3,0x29,0x8F,0x3D,0x35,0xCB};
*/

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))

/*****************************************************************************/
static int gostc_par512(const void* params) 
{
	uint32_t *p = (uint32_t*)params;

	if (!p) return 0;

	/* Признак того, что это 64-разр. набор ГОСТ2012 */
	if((p[0] == 0x28be6373) || (p[0] == 0xFFFFFDC7) || (p[0] == 0x0000006F)) 
		return 1;
	return 0;
}
/*****************************************************************************/
/********************************************************************************/
/*Процедуры вычисления подписи MakeSign и MakeSign64   	
(MakeSign64 вызывается из MakeSign если набор 512-битный).
par - контекст, rnd(32)64b - случайное заполнение для вычисления ЭЦП,
h - хэш-сообщение, на которое вычисляется подпись,
keyD - закрытый ключ вычисления подписи,
sign - буфер для сохранения подписи						*/
/*Возврат: 1 - подпись вычислена успешно, 0 - ошибка вычисления			*/
/*2 - требуется другая случайная посл-ть					*/
static int gostc_MakeSign64(const gostc_CR_SYSTEM64* par, uint8_t* rnd64b,void *h, void *keyD,void *sign)
{	
	uint32_t e[(ASYM_DEF_512_SIZE>>2)+1];
	uint32_t Qmod[(ASYM_DEF_512_SIZE>>2)+1];
	uint32_t d[ASYM_DEF_512_SIZE>>2];
	uint32_t zero[ASYM_DEF_512_SIZE>>2];
	uint32_t r[ASYM_DEF_512_SIZE>>2];
	uint32_t rb[(ASYM_DEF_512_SIZE>>2) + 1];
	uint32_t rd[ASYM_DEF_512_SIZE>>2];
	uint32_t ke[ASYM_DEF_512_SIZE>>2];
	uint32_t p_s[ASYM_DEF_512_SIZE>>2];
	uint8_t k[ASYM_DEF_512_SIZE];
	gostc_POINT_ECP64 p1;
	/*int i;*/
/*	FILE *f;*/
	int rc = 0;

	if(!par || !keyD || !rnd64b || !sign || !h)
		return 0;

	ZZ(zero);
	ZZ(d);
	ZZ(Qmod);
	ZZ(r);
	ZZ(rb);
	ZZ(rd);
	ZZ(ke);
	ZZ(e);
	ZZS(p1);

	/*f = fopen("rez.txt","a");*/
	memcpy(e,h,ASYM_DEF_512_SIZE);
	memcpy(Qmod,par->Q,ASYM_DEF_512_SIZE);

	if(gostc_Divide_c16(e,Qmod,p1.BUF,ASYM_DEF_512_SIZE>>2)==0)
		goto end;/*возврат 0*/
	if(memcmp(zero,e,ASYM_DEF_512_SIZE)==0)
		e[0] = 1;

	memcpy(k,rnd64b,ASYM_DEF_512_SIZE);

	if(memcmp(zero,k,ASYM_DEF_512_SIZE)==0) 
		{rc=2; goto end;} /*Cлучайная последовательность не д.б. нулевой, здесь rc=2*/	

	if (gostc_Comp_C(par->Q,(uint32_t*)k,ASYM_DEF_512_SIZE>>2)!=1) 
		{ rc = 2; goto end; }/*И здесь тоже - она не приведена по модулю*/

	if(gostc_FindKrPoint64(par,k,0, &p1)<0)
		goto end;/*rc=0*/


	memcpy(rb,p1.xp,ASYM_DEF_512_SIZE);
	ZZ(p1.BUF);
	if(gostc_Divide_c16((uint32_t*)rb,Qmod,(uint32_t*)p1.BUF,ASYM_DEF_512_SIZE>>2)==0)
		goto end;/*rc=0*/
	
	memcpy(r,rb,ASYM_DEF_512_SIZE);
	memcpy(d,keyD,ASYM_DEF_512_SIZE);
	if(gostc_Find16ModP(r,d,Qmod,rd)<0)
		goto end; /*rc=0*/
	
	if(gostc_Find16ModP((uint32_t*)k,e,Qmod,ke)<0)
		goto end; /*rc=0*/
	ZZ(p_s);
	if(gostc_AplusBModP64(rd,ke,Qmod,p_s)<0)
		goto end; /*rc=0*/
	memcpy(sign,r,ASYM_DEF_512_SIZE);
	memcpy((uint8_t*)sign+ASYM_DEF_512_SIZE,p_s,ASYM_DEF_512_SIZE);
	rc = 1;/*ОК*/
end:
	SZZS(p1);
	SZZ(e);
	SZZ(Qmod);
	SZZ(d);
	SZZ(r);
	SZZ(rb);
	SZZ(rd);
	SZZ(ke);
	SZZ(p_s);
	SZZ(k);

	return rc;
}

/*################################################################################*/
int gostc_asym_MakeSign(void* params, uint8_t* rnd32b, void *h, void *keyD, void *sign)
{
	const gostc_CR_SYSTEM* par = (gostc_CR_SYSTEM*)params;
	gostc_POINT_ECP p1; 
	uint32_t zero[ASYM_DEF_SIZE>>2];
	uint32_t e[(ASYM_DEF_SIZE>>2)+1];
	uint8_t k[ASYM_DEF_SIZE];
	uint32_t r[ASYM_DEF_SIZE>>2];
	uint32_t rb[(ASYM_DEF_SIZE>>2)+1];
	uint32_t d[ASYM_DEF_SIZE>>2];
	uint32_t rd[ASYM_DEF_SIZE>>2];
	uint32_t ke[ASYM_DEF_SIZE>>2];
	uint32_t prom_s[ASYM_DEF_SIZE>>2];
	uint32_t divisor[(ASYM_DEF_SIZE>>2)+1];
	int a=0;
	int rc = 0,ret=0;

	if(!par || !keyD || !rnd32b || !sign || !h)
		return 0;


	if (gostc_par512(params)) {
		rc = gostc_MakeSign64((const gostc_CR_SYSTEM64*)params,rnd32b,h,keyD,sign);
		return rc;
	}

	ZZ(zero);
	ZZ(e);
	ZZ(k);
	ZZ(r);
	ZZ(rb);
	ZZ(d);
	ZZ(rd);
	ZZ(ke);
	ZZ(prom_s);
	ZZ(divisor);
	ZZS(p1);

	a = gostc_Comp_C((uint32_t*)zero,(uint32_t*)keyD,ASYM_DEF_SIZE>>2);
	if (a == 2|| a <0) 
	{
		rc = 0;
		goto end;
	}
	
	memcpy(e,h,ASYM_DEF_SIZE);
	memcpy(divisor, par->Q, ASYM_DEF_SIZE);

	ZZ(p1.BUF);

	if(gostc_Divide_c((uint32_t*)e,divisor,(uint32_t*)p1.BUF,ASYM_DEF_SIZE>>2)==0)
		goto end;/*rc=0*/
	
	
	if(memcmp(zero,e,ASYM_DEF_SIZE)==0)
		e[0] = 1;

	memcpy(k,rnd32b,ASYM_DEF_SIZE);

	if(memcmp(zero,k,ASYM_DEF_SIZE)==0) 
	{
		rc=2; 
		goto end;
	} /*Cлучайная последовательность не д.б. нулевой, здесь rc=2*/	

	if (gostc_Comp_C((uint32_t*)par->Q,(uint32_t*)k,ASYM_DEF_SIZE>>2)!=1)
	{
		rc = 2;/*Ошибка случайного заполнения: rc=2*/
		goto end;
	}

	ret = gostc_FindKrPointInM(par, k, 0, &p1);
	if(ret < 0)/*Ошибка вычисления кратной точки!!!*/
		goto end;/*rc=0*/

		
	memcpy(rb,p1.xp,ASYM_DEF_SIZE);
	ZZ(p1.BUF);
	memcpy(divisor, par->Q, ASYM_DEF_SIZE);
	if(gostc_Divide_c((uint32_t*)rb, divisor, (uint32_t*)p1.BUF,ASYM_DEF_SIZE>>2)==0)
		goto end;/*rc=0*/
	memcpy(r,rb,ASYM_DEF_SIZE);
	memcpy(d,keyD,ASYM_DEF_SIZE);
	if(gostc_FindEQABModP(r,d,par->Q,rd)<0)
		goto end;/*rc=0*/
	if(gostc_FindEQABModP((uint32_t*)k,e,par->Q,ke)<0)
		goto end;
	/*rc=0*/
	if(gostc_AplusBModP_C(rd,ke,par->Q,prom_s)<0)
		goto end;/*rc=0*/
	
	memcpy(sign,r,ASYM_DEF_SIZE);
	memcpy((uint8_t*)sign+ASYM_DEF_SIZE,prom_s,ASYM_DEF_SIZE);
	rc = 1;/*ОК*/
end:
	SZZS(p1);
	SZZ(e);
	SZZ(k);
	SZZ(r);
	SZZ(rb);
	SZZ(d);
	SZZ(rd);
	SZZ(ke);
	SZZ(prom_s);
	SZZ(divisor);

	return rc;
}
/************************************************************************/
/*********************Проверка подписи (512-битные наборы параметров)****/
/*Вызывается из gostc_CheckSign						*/
/*Возврат: 0 - Ошибка проверки ЭЦП, >1 - ошибка вычислений или передачи параметров
1 - ОК									*/
static int gostc_CheckSign64(const gostc_CR_SYSTEM64* params, void *keyQ, void *h, void *sign)
{
	uint32_t zero[ASYM_DEF_512_SIZE>>1],z1[ASYM_DEF_512_SIZE>>2],z2[ASYM_DEF_512_SIZE>>2],rv[ASYM_DEF_512_SIZE>>2],
Xrez[(ASYM_DEF_512_SIZE>>2)+1],ap[(ASYM_DEF_512_SIZE>>2)+1],Q[(ASYM_DEF_512_SIZE>>2)+1],e[(ASYM_DEF_512_SIZE>>2)+1]
,gamma[ASYM_DEF_512_SIZE>>2],s2[ASYM_DEF_512_SIZE>>2];
	int a=0;
	int rc = 0;
	gostc_POINT_ECP64 p1;
	gostc_POINT64 pQ;	

	if((params == NULL)||(keyQ == NULL)||(h == NULL)||(sign == NULL))
		return 2;

	ZZ(zero);
	ZZ(ap);
	ZZ(e);
	ZZ(s2);
	ZZ(gamma);
	ZZ(Q);
	ZZ(z1);
	ZZ(z2);
	ZZ(rv);
	ZZ(Xrez);
	ZZS(p1);
	ZZS(pQ);

	a = gostc_Comp_C(zero,keyQ,(ASYM_DEF_512_SIZE>>1));
	if (a == 2|| a <0) { rc = 2; goto end; }
	a = gostc_Comp_C(sign,zero,(ASYM_DEF_512_SIZE>>1));
	if(a == 2 || a<0 ) { rc = 2; goto end; }/*/////Ошибки передачи параметров*/

	memcpy(pQ.X,keyQ,ASYM_DEF_512_SIZE);
	memcpy(pQ.Y,(uint32_t*)keyQ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);/*////Скопировали открытый ключ в виде координат второй точки.*/
/*	f1 = fopen("checksgn.txt","a");*/
	
/*	////Переведем координаты второй точки в представление Монтгомери*/
	if(gostc_Find16ModPinM(params,pQ.X,params->R,params->P,pQ.X)<0)
	{
		rc=2;
		goto end;
	}	
	if(gostc_Find16ModPinM(params,pQ.Y,params->R,params->P,pQ.Y)<0)
	{
		rc=2;
		goto end;
	}
	memcpy(pQ.Z,params->M1,ASYM_DEF_512_SIZE); /*////Зарядим единицу в проективных координатах*/
	memcpy(ap,h,ASYM_DEF_512_SIZE);
	memcpy(Q,params->Q,ASYM_DEF_512_SIZE);

	if(gostc_Divide_c16(ap,Q,p1.BUF,ASYM_DEF_512_SIZE>>2)==0)
	{
		rc=2;
		goto end;
	}
	memcpy(e,ap,ASYM_DEF_512_SIZE);
	
		/*//е вычислили.*/
	if(memcmp(zero,e,ASYM_DEF_512_SIZE)==0)
		e[0] = 1;

	memcpy(s2,(uint8_t*)sign+ASYM_DEF_512_SIZE,ASYM_DEF_512_SIZE);
	
	if(gostc_BinaryEuclid64(e,Q,gamma)<0)/*////Нашли гамму = 1/е mod Q*/
	{
		rc = 2;
		goto end;
	}

	if(gostc_Find16ModP(gamma,s2,Q,z1)<0)/*////Z1 вычислили*/
	{
		rc =2;
		goto end;
	}

	if(gostc_Find16ModP(sign,gamma,Q,rv)<0)
	{
		rc=2;
		goto end;
	}

	if(gostc_Sub_C64(Q,rv,z2)<0)
	{
		rc=2;
		goto end;
	}
	
	if(gostc_FindSumKrPoints64(params,&pQ, (uint8_t*)z1,(uint8_t*)z2, Xrez)<0)
	{
		rc = 2;/*Ошибка вычисления суммы кратных точек!*/
		goto end;
	}
	
	ZZ(p1.BUF);
	if(gostc_Divide_c16(Xrez,Q,p1.BUF,ASYM_DEF_512_SIZE>>2)==0)
	{
		rc=2;
		goto end;
	}

	if(memcmp(Xrez,sign,ASYM_DEF_512_SIZE)) 
	{
		/*//ЭЦП не сошлось: rc=0*/
		rc = 0;		
		goto end; 
	}
	/*//Правильно*/
	rc = 1;
end:

	return rc;
}
/*********************Проверка подписи (256-битные наборы параметров)****/
int gostc_asym_CheckSign(void* par, uint8_t* keyQ, uint8_t* hash, uint8_t *sign)
{
	const gostc_CR_SYSTEM *params = (gostc_CR_SYSTEM*)par;
	uint32_t zero[ASYM_DEF_SIZE>>1];
	uint32_t ap[(ASYM_DEF_SIZE>>2)+1];
	uint32_t e[ASYM_DEF_SIZE>>2];
	uint32_t s2[ASYM_DEF_SIZE>>2];
	uint32_t gamma[ASYM_DEF_SIZE>>2];
	/*uint32_t gamma2[8] = {0};//gamma2 для отладки */
	uint32_t q[ASYM_DEF_SIZE>>2];
	uint32_t z1[ASYM_DEF_SIZE>>2];
	uint32_t z2[ASYM_DEF_SIZE>>2];
	uint32_t rv[ASYM_DEF_SIZE>>2];
	uint32_t Xrez[(ASYM_DEF_SIZE>>2)+1];
	gostc_POINT_ECP p1; 
	gostc_ONLY_POINT pQ;
	int a=0;
	int rc = 0;

	if((params == NULL)||(keyQ == NULL)||(hash == NULL)||(sign == NULL))
		return 2;
	ZZ(zero);
	ZZ(ap);
	ZZ(e);
	ZZ(s2);
	ZZ(gamma);
	ZZ(q);
	ZZ(z1);
	ZZ(z2);
	ZZ(rv);
	ZZ(Xrez);
	ZZS(p1);
	ZZS(pQ);

	if (gostc_par512(params)) {
		rc = gostc_CheckSign64((const gostc_CR_SYSTEM64*)par,keyQ,hash,sign);
		return rc;
	}

	a = gostc_Comp_C(zero,(uint32_t*)keyQ,ASYM_DEF_SIZE>>1);
	if (a == 2 || a <0) {
		rc = 2;
		goto end;
	}
	a = gostc_Comp_C((uint32_t*)sign,zero,ASYM_DEF_SIZE>>1);
	if(a == 2 || a<0 ) {
		rc = 2;
		goto end;
	}

	memcpy(pQ.X,keyQ,ASYM_DEF_SIZE);
	memcpy(pQ.Y,(uint8_t*)keyQ+ASYM_DEF_SIZE,ASYM_DEF_SIZE);
	
	if(gostc_FindEQABModPinM(params,pQ.X,params->R,params->P,pQ.X)<0)
	{
		rc=2;
		goto end;
	}
	if(gostc_FindEQABModPinM(params,pQ.Y,params->R,params->P,pQ.Y)<0)
	{
		rc=2;
		goto end;
	}
	memcpy(pQ.Z,params->M1,ASYM_DEF_SIZE);/*Перевод в проективные координаты(x,y) -> (X,Y,1)*/

	memcpy(ap,hash,ASYM_DEF_SIZE);
	ZZ(p1.BUF);

	if(gostc_Divide_c((uint32_t*)ap,(uint32_t*)params->Q,(uint32_t*)p1.BUF,ASYM_DEF_SIZE>>2)==0)
	{
		rc=2;
		goto end;
	}
	memcpy(e,ap,ASYM_DEF_SIZE);
		/* е вычислили. */
	if(memcmp(zero,e,ASYM_DEF_SIZE)==0)
		e[0] = 1;
	memcpy(s2,sign+ASYM_DEF_SIZE,ASYM_DEF_SIZE);
	memcpy(q,params->Q,ASYM_DEF_SIZE);
/*	BinaryEuclid(e,q,gamma2,&params->p1); */
	if(gostc_BinaryEuclid_C(e,q,gamma)<0)
	{
		rc = 2;
		goto end;
	}
	if(gostc_FindEQABModP(gamma,s2,q,z1)<0)
	{	rc=2;
		goto end;
	}
	if(gostc_FindEQABModP((uint32_t*)sign,gamma,q,rv)<0)
	{
		rc=2;
		goto end;
	}
	if(gostc_Sub_C(q,rv,z2)<0)
	{
		rc=2;
		goto end;
	}
	if(gostc_FindSumOfKrPoints(params,&pQ, z1,z2,Xrez)<0)
	{
		rc = 2;/*Ошибка вычисления суммы кратных точек*/
		goto end;
	}
	ZZ(p1.BUF);
	if(gostc_Divide_c((uint32_t*)Xrez,(uint32_t*)q,(uint32_t*)p1.BUF,ASYM_DEF_SIZE>>2)==0)
	{
		rc=2;
		goto end;
	}
	if(memcmp(Xrez,sign,ASYM_DEF_SIZE)==0)
	{
		rc = 1; /* Правильно */
		goto end;
	}
	rc = 0;/* ЭЦП не сошлось */
end:

	return rc;
}

/****************************************************************************/
/*Процедура заполнения контекста base эталонными значениями, а также рассчета промежуточных наполнений контекста
(используются при вызове gostc_asym_LoadParams), для 256-битных наборов*/
static int gostc_asym_InitializeSystem(void *to, /*void *RandomFunc,*/ void *from)
{
	int i=0;
	uint32_t *bK = NULL;
	uint32_t *fr = NULL;
	gostc_CR_SYSTEM *base = (gostc_CR_SYSTEM*)to;

	if(!to || !from)
		return -1;/*Ошибка передачи параметров*/	

	memcpy(base,from, ASYM_BASE_OFFSET);
	
	/* Загружаем кратные точки */
	bK = (uint32_t*)base +  (ASYM_BASE_OFFSET>>2);
	fr = (uint32_t*)from +  (ASYM_BASE_OFFSET>>2);
	
	for(i=0;i<ASYM_COUNTED_KPOINTS;i++)
	{
		memcpy(bK,fr,ASYM_POINTSZ_256BIT*sizeof(uint32_t));
		bK = bK + ASYM_POINTSZ_256BIT;
		fr = fr + ASYM_POINTSZ_256BIT;
	}
	bzero(&base->R,ASYM_DEF_SIZE);
	bzero(&base->halfP,ASYM_DEF_SIZE);
	bzero(&base->M1,ASYM_DEF_SIZE); /* единица в представлении Монтгомери */
	/*	bzero(&base->prBUF,32);*/
	return 0;/*ОК*/
}
/*****************************************************************************/
/*Процедура заполнения контекста base эталонными значениями, а также рассчета промежуточных наполнений контекста
(используются при вызове gostc_asym_LoadParams), для 512-битных наборов
*/

static int gostc_asym_InitializeSystem64(void *to,void * from)
{
	gostc_CR_SYSTEM64 *par=NULL; 
	gostc_POINT_ECP64 p1;
	uint32_t *source=NULL, *p0=NULL; 
	uint32_t R0[(ASYM_DEF_512_SIZE>>1)+2],m3[ASYM_DEF_512_SIZE>>2],divisor[(ASYM_DEF_512_SIZE>>2)+1];
	int i=0,rc =-1;
	if(!to || !from)
		return rc;/*Ошибка передачи параметров*/	

	bzero(R0,sizeof(R0));
	bzero(m3,sizeof(m3));
	ZZ(divisor);
	ZZS(p1);

	m3[0] = 3; /*/////3*/

	par = (gostc_CR_SYSTEM64*)to;
	source = (uint32_t*)from;
	/*////P*/
	memcpy(par->P,source,ASYM_DEF_512_SIZE);
	memcpy(divisor,par->P,ASYM_DEF_512_SIZE);
		
/*///Q*/
	memcpy(par->Q,source+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
/*///единица в монтгомери*/
	memcpy(par->M1,source+(ASYM_DEF_512_SIZE>>2)*2,ASYM_DEF_512_SIZE);
/*  ///  */
	memcpy(par->A,source+(ASYM_DEF_512_SIZE>>2)*3,ASYM_DEF_512_SIZE);
/* /////X0,Y0 */
	memcpy(par->X0,source+(ASYM_DEF_512_SIZE>>2)*4,ASYM_DEF_512_SIZE);
	memcpy(par->Y0,source+(ASYM_DEF_512_SIZE>>2)*5,ASYM_DEF_512_SIZE);
	
	
	R0[ASYM_DEF_512_SIZE>>1]=1;
	if(gostc_Divide_c16(R0,divisor,p1.BUF,(ASYM_DEF_512_SIZE>>1)+1)==0)
	{
		rc=-2;
		goto end;
	}
	
	p0 = (uint32_t*)par->P;
	par->EpsP = (uint32_t)gostc_EpsilonQ(*p0);
	if(!par->EpsP)
	{
		rc=-3;
		goto end;
	}
	memcpy(par->R,R0,ASYM_DEF_512_SIZE);

	memcpy(R0,par->P,ASYM_DEF_512_SIZE);
	R0[0]++;
	
	for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
	{
		R0[i] = R0[i]>>1;
		if((R0[i+1]&1) == 1)
			R0[i] = R0[i] + ASYM_ELDER;
	}
	R0[(ASYM_DEF_512_SIZE>>2)-1] = R0[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
	memcpy(par->halfP,R0,ASYM_DEF_512_SIZE);/*///Половинка P*/
	
	if(gostc_Find16ModPinM(par,par->A,par->R,par->P,par->AM)<0)
	{
		rc=-3;
		goto end;
	}
/*	////Заполним точки с монтгомери*/
/*	////Сперва базовая в проект. координатах. */
	ZZS(par->P0);

	memcpy(par->P0.X,par->X0,ASYM_DEF_512_SIZE);
	memcpy(par->P0.Y,par->Y0,ASYM_DEF_512_SIZE);
	par->P0.Z[0] = 1;
/*	////Переведем её в представление Монтгомери*/
	ZZS(par->PM0);
	if(gostc_Find16ModPinM(par,par->P0.X,par->R,par->P,par->PM0.X)<0)
	{
		rc=-3;
		goto end;
	}
	if(gostc_Find16ModPinM(par,par->P0.Y,par->R,par->P,par->PM0.Y)<0)
	{
		rc=-3;
		goto end;
	}
	memcpy(par->PM0.Z,par->M1,ASYM_DEF_512_SIZE);
	
	if(gostc_Find16ModPinM(par,m3,par->R,par->P,par->M3)<0) /*////3 в представлении Монтгомери */
	{
		rc=-3;
		goto end;
	}
	
	
	rc=0;/*OK*/
end:
	SZZ(R0);
	SZZ(m3);
	SZZ(divisor);
	SZZS(p1);
	return rc;
}
/******************************************************************************/


/*****************************************************************************/
void *gostc_asym_LoadParams(int n)
{
	gostc_CR_SYSTEM *params = NULL;
	gostc_CR_SYSTEM64 *p2 = NULL;
	gostc_POINT_ECP p1; 
	uint32_t R0[(ASYM_DEF_SIZE>>1)+2];
	uint32_t *p0 = NULL;
	uint32_t divisor[(ASYM_DEF_SIZE>>2)+1];
	int i=0,rc=0;

	if((n < 1) || (n > 7)) {
		return NULL;
	}

	if(n >= 5 && n <= 7)
	{
		p2 = gostc_malloc(sizeof(gostc_CR_SYSTEM64));
		if(p2 == NULL)
		{
			return NULL;	
		}
		if(n == 5)
		{	if(gostc_asym_InitializeSystem64(p2,gostc_asym_64set)<0)
			{	gostc_free(p2);
				return NULL;
			}
			else
				return p2;
		}
		if(n == 6)
		{
			if(gostc_asym_InitializeSystem64(p2,gostc_asym_GOST64ParamSetA)<0)
			{	gostc_free(p2);
				return NULL;
			}
			else
				return p2;
		}
		if(n == 7)
		{
			if(gostc_asym_InitializeSystem64(p2,gostc_asym_GOST64ParamSetB)<0)
			{	gostc_free(p2);
				return NULL;
			}			
			else
				return p2;
		}
	}
	ZZ(R0);
	ZZ(divisor);
	ZZS(p1);
	params = gostc_malloc(sizeof(gostc_CR_SYSTEM));
	if(!params)
		return NULL;
	
/*	bzero(params,sizeof(gostc_CR_SYSTEM)); - already zeroed by gostc_malloc */
	if(n==1)/* Номер 1 - тестовый набор */
		rc = gostc_asym_InitializeSystem(params, gostc_asym_GOSTset);
	if(n==2)/* Номер 2 - CSPa */
		rc = gostc_asym_InitializeSystem(params, gostc_asym_CSPset);
	if(n==3)/* Номер 3 - OSCAR */
		rc = gostc_asym_InitializeSystem(params, gostc_asym_OSset);
	if(n==4)/* Номер 4 - CSPb */
		rc = gostc_asym_InitializeSystem(params, gostc_asym_CSP2set);
	if(rc < 0)/*Ошибка!*/
	{
		gostc_free(params);
		return NULL;
	}

	memcpy(divisor, params->P,ASYM_DEF_SIZE);
	R0[(ASYM_DEF_SIZE>>1)] = 1;
	if(gostc_Divide_c((uint32_t*)R0,(uint32_t*)divisor,(uint32_t*)p1.BUF,(ASYM_DEF_SIZE>>1)+1)==0)/*R = R0 % P*/
	{
		gostc_free(params);
		return NULL;
	}
	p0 = (uint32_t*)divisor;
	params->EpsP = gostc_EpsilonQ(*p0);
	if(!params->EpsP)
	{
		gostc_free(params);
		return NULL;	
	}
	memcpy(params->R,R0,ASYM_DEF_SIZE);
	memcpy(R0,params->P,ASYM_DEF_SIZE);
	R0[0]++;
	
	for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
	{
		R0[i] = R0[i]>>1;
		if(R0[i+1]%2 == 1)
			R0[i] = R0[i] + ASYM_ELDER;
	}
	R0[(ASYM_DEF_SIZE>>2)-1] = R0[(ASYM_DEF_SIZE>>2)-1] >> 1;
	
	
	memcpy(params->halfP,R0,ASYM_DEF_SIZE);
	memcpy(params->M1,params->PM0.Z,ASYM_DEF_SIZE);/* сохраняем единицу в представлении Монтгомери */
	
	if(gostc_FindEQABModPinM(params,params->A,params->R,params->P,params->AM)<0)
	{
		gostc_free(params);
		return NULL;
	}
	return params;
}
/*****************************************************************************/

/*****************************************************************************/
void gostc_asym_FreeParams(void *params)
{
	if (!params)
		return;
	gostc_free(params);
}
/*****************************************************************************/
/* ////////Генерация ключей////////////////////////// */
int gostc_asym_GenSecretKey(void* params, uint8_t* rnd32b, void* keyD)
{
	const gostc_CR_SYSTEM64* p2;
	const gostc_CR_SYSTEM* par = (const gostc_CR_SYSTEM*)params;

	if(!params || !rnd32b || !keyD)
		return 0;
	
	if (gostc_par512(params)) 
	{
		p2 = (const gostc_CR_SYSTEM64*)params;
		if (gostc_Comp_C((uint32_t*)p2->Q,(uint32_t*)rnd32b,ASYM_DEF_512_SIZE>>2)!=1)
			return 2; /* Incorrect Random  */
		else
		{
			memcpy(keyD, rnd32b, ASYM_DEF_512_SIZE);/*64 байта!*/
			return 1;/*OK*/
		}
	}
	if (gostc_Comp_C((uint32_t*)par->Q,(uint32_t*)rnd32b,ASYM_DEF_SIZE>>2)!=1)
		return 2; /* Incorrect Random */
	memcpy(keyD, rnd32b, ASYM_DEF_SIZE);
	return 1;/*OK*/
}
/********************************************************************************/
/*Генерация открытого ключа для 512- битного набора
вызывается из gostc_GenPublicKey				*/
static int gostc_GenPublicKey64(const gostc_CR_SYSTEM64* par,void* keyD,void* keyQ)
{
	uint8_t kkey[ASYM_DEF_512_SIZE]; 
	uint32_t zero[ASYM_DEF_512_SIZE>>2];
	gostc_POINT_ECP64 p1;
	int a=0;
	int rc = 0;
	
	if(par == NULL||keyD == NULL || keyQ == NULL)
		goto end;/*Ошибка передачи параметров*/

	ZZ(zero);
	ZZS(p1);


	memcpy(kkey,keyD,ASYM_DEF_512_SIZE);
	
	
	a = gostc_Comp_C(zero,keyD,ASYM_DEF_512_SIZE>>2);
	
	if (a == 2 || a <0 ) goto end; /*/////Закрытый ключ некорректен!*/
	
	if(gostc_FindKrPoint64(par,kkey,0,&p1)<0) goto end;/*Ошибка вычисления открытого ключа!*/

	memcpy(keyQ,p1.xp,ASYM_DEF_512_SIZE);
	memcpy((uint8_t*)keyQ+ASYM_DEF_512_SIZE,p1.yp,ASYM_DEF_512_SIZE);

	rc = 1;
end:
	SZZ(kkey);
	SZZS(a);
	SZZS(p1);

	return rc;
}
/*Генерация открытого ключа keyQ = keyD*P */
int gostc_asym_GenPublicKey(void* params, void* keyD, void* keyQ)
{
	const gostc_CR_SYSTEM* par = (const gostc_CR_SYSTEM*)params;
	uint8_t kkey[ASYM_DEF_SIZE];
	uint32_t zero[ASYM_DEF_SIZE>>2];
	uint32_t tst[ASYM_DEF_SIZE>>2];
	gostc_POINT_ECP p1;
	int a=0;
	int rc = 0;

	if(params == NULL||keyD == NULL || keyQ == NULL)
		return rc;/*Ошибка передачи параметров*/

	if (gostc_par512(params)) 
	{
		rc = gostc_GenPublicKey64((const gostc_CR_SYSTEM64*)params,keyD,keyQ);
		return rc; /*////Ошибка генерации - возврат 0 при ошибке,1 - ОК*/
	}


	ZZ(zero);
	ZZ(tst);
	ZZS(p1);

	memcpy(kkey, keyD, ASYM_DEF_SIZE);

	a = gostc_Comp_C(zero,keyD,ASYM_DEF_SIZE>>2);
	if (a == 2 || a <0 )
		goto end;
	
	if(gostc_FindKrPointInM(par,kkey, 0,&p1)<0)
		goto end;
	

	if(gostc_FindEQABModP(keyD,p1.xp,par->P,tst)<0)
		goto end;
	
	
	memcpy(keyQ,p1.xp,ASYM_DEF_SIZE);
	memcpy((uint8_t*)keyQ+ASYM_DEF_SIZE,p1.yp,ASYM_DEF_SIZE);
	rc = 1;/*ОК*/
end:
	SZZ(kkey);/*Зануляем закрытый ключ*/
	SZZS(p1);
	SZZ(tst);
	
	return rc;
}
/*****************************************************************************/
int gostc_asym_GenDFHKey(void* params, void* keyD, void* keyQ, void* UKM, void* keyDFH)
{
	if (!params) return 0;
	if (gostc_par512(params)) 
		return gostc_asym_GenDFHKey2(params, keyD, keyQ, UKM, ASYM_DEF_512_SIZE>>2, keyDFH);
	
	return gostc_asym_GenDFHKey2(params, keyD, keyQ, UKM, ASYM_DEF_SIZE>>2, keyDFH);
}
/*****************************************************************************/
int gostc_asym_GenDFHKey2(void* params, void* keyD, void* keyQ, 
	void* ukm, size_t ukm_len, void* keyDFH)
{
	uint32_t ukm16[ASYM_UKM16_SIZE>>2];
	int rc = 0;

	if(!params || !keyD || !keyQ || !ukm) return 0;

	if (ukm_len == ASYM_DEF_512_SIZE>>2) memcpy(ukm16, ukm, ASYM_DEF_512_SIZE>>2);
	else if (ukm_len == ASYM_DEF_SIZE>>2) { ZZ(ukm16); memcpy(ukm16, ukm, ASYM_DEF_SIZE>>2); }
	else return 0;
	
	if (gostc_par512(params))
	{
		rc = gostc_GenDFHKey64((const gostc_CR_SYSTEM64*)params,keyD, ukm16, keyQ,keyDFH);
		return rc;
	}

	/* 32-разр */
	rc = gostc_GenDFHKey32((const gostc_CR_SYSTEM*)params, keyD, ukm16, keyQ, keyDFH);
	return rc;
}
/*****************************************************************************/
/* Наборы для тестирования ЭЦП */
uint32_t gostc_kTest[] = 
{0xCE5BBB62,0x3475122F,0x94128B9F,0x4FC6AFC2,
0xA61EE031,0x3FA4D560,0x43B22A6F,0x769C1995};

uint32_t gostc_hTest[] = {0xBEE060DB,0x43E7CC88,0x24F23DD7,0xD1E7FAD2,
	0x92AEF5E4,0x224141CB,0x8690AC02,0x54B6DD0A
};

uint32_t gostc_skTest[] = {0x717DDBE6,0xC7ED4674,0x82B78342,0x0C628E60,
	0x78E88CEB,0xCD9CBE43,0x48763A2F,0xF5F3F8A0
};

uint32_t gostc_pkTestX[] = {
	0x58f979a9,0xfa87a53e,0x095898bc,0x06c2cba9,
	0xc6e3a32c,0x8eb16f53,0x8ea2a1d7,0x0de53df4
};

uint32_t gostc_pkTestY[] = {
	0x44f6aa2a,0x2aa3e9d5,0x1e0d8423,0x30c84f9d,
	0x72e0975d,0xe615319e,0x2117a510,0xdacc2358
};

uint32_t gostc_sgnTest[] = {
	0x88957936,0x64b67aea,0xb958242e,0xbe864b39,
	0x919ce60f,0xe25f684a,0x5a2307b0,0xd3393049,
	0x527628ed,0x4459e857,0xa0de4160,0xf52ffa60,
	0x2dbdf532,0x7853cbeb,0xb4f6fe06,0xbb8fc07e
};

/**********************64-разрядный набор******************************/
uint32_t gostc_skTest16[] = {
	0x1fa18dd4,0xc6296782,0xfd18aadf,0xa2636b7b,0x827e2714,0x3322dad2,0x2200a056,0x72e8123b,
	0x508b1020,0x0ee7508e,0x51466970,0x3091a0e8,0x7756d7c9,0xa40936d4,0xadae241b,0x0ba6048a
};

uint32_t gostc_pkTestX16[] = {
	0xd530efe1,0xdd33612c,0x5c1d9dd9,0xf75c4541,0x4c8b4ddf,0x69bc5b92,0xd13314af,0x5a515856,
	0x854621dd,0x5b5c320c,0xbe33c181,0xc4a85a65,0xb9e740d4,0x48598d8a,0x96760c7b,0x115dc5bc
};

uint32_t gostc_pkTestY16[] = {
0x3677beec,0x7f35eca9,0x9339fdf2,0x4c114e1f,0x59a3cdb8,0xf0c70a27,0xd943ffe7,0xea199441,
0xb72afd61,0x639f5d7f,0x503b3d52,0xe2634fa0,0xb751cfa0,0x21dc3ac1,0xd40b0f56,0x37c7c90c
};
uint32_t  gostc_kTest16[] =
	{0x1ae679f1,0xa3af71bb,0xd14cf70e,0x212273a6,0x11842286,0x44340060,0xa44b3e79,0x86748ed7,
	0x364f3658,0xd455986e,0x0b39d019,0x94631212,0xc6801496,0xcc570456,0xb1410fea,0x0359e7f4};

uint32_t gostc_hTest16[] =
	{0x72075b8c,0xc6777d29,0x1d560c4f,0x407adedb,0x647c5d5a,0x4339976c,0x6593f441,0x7184ee53,
	0x35b0c591,0xa71d1470,0xc170c533,0x1b09b6f9,0x4d8dab53, 0x5c4f4a7c,0xacc9e061,0x3754f3cf
};

uint32_t gostc_sgnTest16[] = 
{
0xe173ae36,0x17e19344,0xcd9c5c33,0x60c93bcb,0x06998502,0x9ec197c9,0x86b20f1c,0x54925584,
0xa8fcacd3,0x643c78ee,0x2ee0dcc2,0xe512a3c8,0x1e3c689e,0x23dd795e,0xa081091a,0x2f86fa60,
0xe73e5b4a,0x2a9853bd,0x56919cb9,0x6a6eeb1f,0x7f70ce40,0x526080df,0xe8c4f362,0x823ce288,
0x33d52ff5,0x1c0bfbe9,0x8aadbc08,0x325f5677,0x362d26b6,0x6585e7a9,0x696ffe8e,0x1081b394
};
/*
void _cdecl TestRand(void * where, int n)
{
	memcpy(where,kTest,n);
}
*/

/*****************************************************************************/


static int gostc_TestKeyDFH64(const gostc_CR_SYSTEM64* params)
{
	uint32_t sk1[ASYM_DEF_512_SIZE>>2],sk2[ASYM_DEF_512_SIZE>>2],
pk1[ASYM_DEF_512_SIZE>>1],pk2[ASYM_DEF_512_SIZE>>1],dfh1[ASYM_DEF_512_SIZE>>1],dfh2[ASYM_DEF_512_SIZE>>1],
rnd[ASYM_DEF_512_SIZE>>2];
	uint32_t ukm[ASYM_DEF_512_SIZE>>2];
	int i=0,rc=0;
	
	ZZ(sk1);
	ZZ(sk2);
	ZZ(pk1);
	ZZ(pk2);
	ZZ(dfh1);
	ZZ(dfh2);
	ZZ(ukm);
	
	if(!params)
	    return 0;
	
	ukm[0] = 2;
	ukm[1] = 2;
	srand(0);
reptk1:
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
		rnd[i] = (rand()<<16) + rand();
	rc=gostc_asym_GenSecretKey((void*)params, (uint8_t*)rnd, sk1);
	if(rc == 2)/*random - ?*/
		goto reptk1;
	if(!rc)
		return 0;
	
reptk2:
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
		rnd[i] = (rand()<<16) + rand();
	rc=gostc_asym_GenSecretKey((void*)params, (uint8_t*)rnd, sk2);
	if(rc == 2)/*random - ?*/
		goto reptk2;
	if(!rc)
		return 0;
	
/*	memcpy(sk1, gostc_skTest16,64);
	memcpy(sk2,gostc_kTest16,64);*/
/*Сгенерили секретные ключи. Теперь делаем открытые */
    	gostc_asym_GenPublicKey((void*)params, sk1, pk1);
    	gostc_asym_GenPublicKey((void*)params, sk2, pk2);
	
	/*Теперь вычислим ключи диффи-хеллмана и проверим их */
	
	gostc_asym_GenDFHKey((void*)params,sk1,pk2,ukm,dfh1);
	
	gostc_asym_GenDFHKey((void*)params,sk2,pk1,ukm,dfh2);
	if(memcmp(dfh1,dfh2,ASYM_DEF_512_SIZE*2)!=0)
	        return 0;
	
	return 1;
}

int gostc_asym_TestLib()
{
	const gostc_CR_SYSTEM* params = gostc_asym_LoadParams(SIGN_CPROA);
	const gostc_CR_SYSTEM64* p2=NULL;
	uint32_t pkTestX1[ASYM_DEF_SIZE>>1],pkTest16[ASYM_DEF_512_SIZE>>1];
	uint32_t sgn[ASYM_DEF_SIZE>>1],sgn16[ASYM_DEF_512_SIZE>>1];
	/*FILE *f;
	int i;*/
	gostc_asym_GenPublicKey((void*)params, gostc_skTest, pkTestX1);

	
	
	if(memcmp(pkTestX1, gostc_pkTestX,ASYM_DEF_SIZE)!=0)
	{
		gostc_asym_FreeParams((void*)params);
		return 0;
	}

	if(memcmp(pkTestX1+(ASYM_DEF_SIZE>>2), gostc_pkTestY, ASYM_DEF_SIZE)!=0)
	{
		gostc_asym_FreeParams((void*)params);
		return 0;
	}
	
	gostc_asym_MakeSign((void*)params,(uint8_t*)gostc_kTest,gostc_hTest,gostc_skTest,sgn);
	if(memcmp(sgn,gostc_sgnTest,sizeof(sgn))!=0)
	{
		gostc_asym_FreeParams((void*)params);
		return 0;
	}

	if(!gostc_asym_CheckSign((void*)params, (uint8_t*)pkTestX1, (uint8_t*)gostc_hTest, 
		(uint8_t*)sgn))
	{
		gostc_asym_FreeParams((void*)params);
		return 0;
	}

	gostc_asym_FreeParams((void*)params);
	/***********64-хбайтовый набор ГОСТ 2012****************************************/
	p2 = gostc_asym_LoadParams(SIGN_TST512);
	if(!p2)
	{
		return 0;
	}
	/*f = fopen("testsimple.txt","a");*/
	
	gostc_asym_GenPublicKey((void*)p2, gostc_skTest16, pkTest16);
/*	fprintf(f,"pubkey:\n");
	for(i=0;i<32;i++)
		fprintf(f,"0x%8.8x\n",pkTest16[i]);*/
	
	if(memcmp(pkTest16, gostc_pkTestX16,ASYM_DEF_512_SIZE)!=0)
	{
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	
	if(memcmp(pkTest16+(ASYM_DEF_512_SIZE>>2), gostc_pkTestY16,ASYM_DEF_512_SIZE)!=0)
	{
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	gostc_asym_MakeSign((void*)p2,(uint8_t*)gostc_kTest16,gostc_hTest16,gostc_skTest16,sgn16);
	
/*	fprintf(f,"makesign:\n");
	for(i=0;i<32;i++)
		fprintf(f,"0x%8.8x\n",sgn16[i]);*/
	
	if(memcmp(sgn16,gostc_sgnTest16,ASYM_DEF_512_SIZE*2)!=0)
	{
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	if(gostc_asym_CheckSign((void*)p2, (uint8_t*)pkTest16, (uint8_t*)gostc_hTest16, 
		(uint8_t*)sgn16)!= 1)
	{
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	if(!gostc_TestKeyDFH64(p2))
		return 0;
	gostc_asym_FreeParams((void*)p2);

	/*gost2012 paramset A */
	p2 = gostc_asym_LoadParams(SIGN_TC26A);
	if(!p2)
	{
		return 0;
	}
	gostc_asym_GenPublicKey((void*)p2, gostc_skTest16, pkTest16);
	ZZ(sgn16);
	gostc_asym_MakeSign((void*)p2,(uint8_t*)gostc_kTest16,gostc_hTest16,gostc_skTest16,sgn16);
	
	if(gostc_asym_CheckSign((void*)p2, (uint8_t*)pkTest16, (uint8_t*)gostc_hTest16, 
		(uint8_t*)sgn16)!= 1)
	{
		
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	if(!gostc_TestKeyDFH64(p2))
		return 0;

	gostc_asym_FreeParams((void*)p2);
	/*gost2012 paramset B */
	p2 = gostc_asym_LoadParams(SIGN_TC26B);
	if(!p2)
	{
		return 0;
	}
	gostc_asym_GenPublicKey((void*)p2, gostc_skTest16, pkTest16);
	ZZ(sgn16);
	gostc_asym_MakeSign((void*)p2,(uint8_t*)gostc_kTest16,gostc_hTest16,gostc_skTest16,sgn16);
	
	if(gostc_asym_CheckSign((void*)p2, (uint8_t*)pkTest16, (uint8_t*)gostc_hTest16, 
		(uint8_t*)sgn16)!= 1)
	{
		
		gostc_asym_FreeParams((void*)p2);
		return 0;
	}
	if(!gostc_TestKeyDFH64(p2))
		return 0;

	gostc_asym_FreeParams((void*)p2);

	return 1;/*OK*/
}
/*****************************************************************************/

