#include "gostc_mem.h"
#include "gostc_common.h"
#include "gostc_asym_elems.h"

#ifdef CHECKDIV
#include <stdio.h> /* CHECKDIV */
#endif


#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))

/*Вычитание: c = a - b (8-разр.)	*/
/*Возврат: 0 - ОК, <0 - ERROR		*/
int gostc_Sub_C(uint32_t *a, const uint32_t *b, uint32_t *c)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	int i=0;
	if(!a || !b || !c)
		return -1;/*parameters - ERROR*/

	/* считаем А - В */
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)a[i];
		p1 = p1 - (uint64_t)b[i];/* Отнимаем разряд B */
		p1 = p1 - p2; /* Учитываем заём с пред. разряда */
		c[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}
	SZZS(p1); 
	SZZS(p2); 
	return 0;
}

/*Вычитание: c = a - b (16-разр.)*/
int gostc_Sub_C64(uint32_t *a, const uint32_t *b, uint32_t *c)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	int i=0;
	if(!a || !b || !c)
		return -1;/*parameters - ERROR*/
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 = (uint64_t)a[i];
		p1 = p1 - (uint64_t)b[i];/*Отнимаем разряд B*/
		p1 = p1 - p2; /*Учитываем заём с пред. разряда*/
		c[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /*Был новый заем - надо его сохранить*/
			p2 = 1;
		else
			p2 = 0;
	}

	SZZS(p1); 
	SZZS(p2); 
	return 0;/*OK*/
}


/*Получение EpsilonQ - шага в представлении числа Монтгомери. Q0 не д.б. равен 0 */
uint32_t  gostc_EpsilonQ(uint32_t Q0)
{
	uint64_t b=0;
	int ind = 1;
	uint32_t u0=0,u1=0,r1=0,r2=0,u2=0,q2=0,t=0,epsQ=0;
	
	if(Q0 == 0)/*Ошибка передачи параметров!*/
		return 0;

	b = 1;
	b <<= 32;
	r1 = (uint32_t)(b % Q0);

	if(r1 == 0)
	{
		u2 = 1;
		goto m2;
	}
	q2 = Q0/r1;
	r2 = Q0%r1;
	ind = -ind;
	if(r2 == 0)
	{
		u2 = (uint32_t)(b / Q0);
		goto m2;
	}

	u0 = 1;
	u1 = (uint32_t)(b / Q0);
m1:
	ind = -ind;
	u2 = u1*q2+u0;
	t = r1;
	r1 = r2;
	Q0 = t;
	q2 = Q0/r1;
	r2 = Q0%r1;
	t = u1;
	u1 = u2;
	u0 = t;

	if(r2!=0)
		goto m1;

m2:
	epsQ = -ind * u2;

	SZZS(b);
	SZZS(ind);
	SZZS(u0);
	SZZS(u1); 
	SZZS(r1); 
	SZZS(r2);
	SZZS(u2);
	SZZS(q2);

	return epsQ;
}

/*********************************************************************************************	*/
/*Целочисленное деление: q = u/v, u = u%v, k1- разрядность делимого. Разрядность делителя: 8 	*/
/*В параметре s1 сохраняются некоторые промежуточные вычисления во время операции. Не должен быть NULL*/
/*(!)Функция возвращает 0 в случае ошибки; 1 в случае успешного завершения операции; 2 - если делитель оказался больше делимого,
в этом случае содержимое делимого является остатком от деления			*/
int gostc_Divide_c(uint32_t *u,uint32_t *v, uint32_t *q, uint32_t k1)
{
	uint32_t /*addr_off1,*/ addr_r_u=0, sd=0, rez_mu=0, rez_mq=0, delitelr=0, div_=0, mod=0;
	uint32_t prev=0;
	uint64_t p1=0, p2=0, delimoer=0, proizv=0, prom=0, up=0;
	int i=0, t=0;
	int rc = 0;
	gostc_BUF_ECP s1={0};

	if((u == NULL)||(v == NULL)||(q == NULL))
		return 0;
	/*addr_off1 = (k1-1)<<2;*/
	addr_r_u = k1;
/* opredelim pervyj znachimyj razrjad. */
	while(addr_r_u > 0 && !u[addr_r_u - 1])
	{	
		addr_r_u--;
	}
	/* catch */
	s1.mu = addr_r_u;/* Sohranili razrjadnost' delimogo */
/*	oldr_u = (uint32_t)u + addr_off1 + 4;
	s1.end_u = oldr_u; * Sohranili adres starshego razrjada. 
*/

	addr_r_u = ASYM_DEF_SIZE>>2;
	while(addr_r_u && !v[addr_r_u - 1])
	{
		addr_r_u--;
	}
	if(addr_r_u == 0) {
		rc = 0;
		goto end; /* Division by zero */
	}

	s1.mv = addr_r_u;
	/* Razobralis' s delitelem 
	 * Opredelim teper' kol-vo neznachawih starshih dvoichnyh razrjadov delitelja	
	 */
	sd = v[s1.mv - 1];/*s1.mv is not 0!!!*/
	i = 31;
	sd = 1 << 31;
	while (!(v[s1.mv - 1]&sd))
	{
		sd = sd>>1;
		i--;
	}
	sd = 31 - i;
	s1.sd = sd; /* Saved */
	
	/* warning! - delitel' po razrjadam bol'she delimogo. */
	if(s1.mv > s1.mu) 
	{
		rc = 2;/*Это не ошибка выполнения операции, след-но rc>0 */ /* было rc = 0;*/
		goto end;
	}
	
	s1.mq = s1.mu - s1.mv + 1; /* Saved mq */
	
	/* Normalizing */
	if(s1.sd!=0)
	{
		p1 = 0;
		p2 = 0;
		for(i = 0; i < (int)(s1.mu - 1); i++)
		{
			p1 = u[s1.mu -1 -i];
			p1 = p1<<s1.sd;
			p2 = u[s1.mu -2 -i];
			p2 = p2<<s1.sd;
			p2 = p2>>32;
			p1 ^= p2;
			u[s1.mu-1-i] = (uint32_t)p1;
		}	
		u[0] = u[0]<<s1.sd; /* Last digit... */
		/* Sdvigaem delitel' */
		for(i = 0; i < (int)(s1.mv - 1); i++)
		{
			p1 = v[s1.mv -1 -i];
			p1 = p1<<s1.sd;
			p2 = v[s1.mv -2 -i];
			p2 = p2<<s1.sd;
			p2 = p2>>32;
			p1 ^= p2;
			v[s1.mv-1-i] = (uint32_t)p1;
		}
		v[0] = v[0]<<s1.sd; /* Last digit... */
		
	}
	rez_mu = s1.mu;
	rez_mq = s1.mq;
mainLoop:
	delimoer = u[s1.mu];
	delitelr = v[s1.mv-1];
	div_ = 0;
	if((uint32_t)delimoer < delitelr)
		goto divide;
	/* Sohranim ASYM_MINUS_1 v starshem razrjade chastnogo */
	q[s1.mq -1] = ASYM_MINUS_1;
	div_ = ASYM_MINUS_1;
	goto subtract;
divide:
	delimoer = (delimoer << 32) + (uint32_t)u[s1.mu -1];
	div_ = (uint32_t)(delimoer/delitelr); /* Quotient */
	mod = delimoer%delitelr; /* Remainder */

	s1.remainder = mod; /* save them */
	q[s1.mq -1] = div_;

	proizv = (uint64_t)(div_) * v[s1.mv-2];
	/* Mnozhim predpolagaemoe chastnoe na sledujuwij razrjad delitelja. */
	if((uint32_t)(proizv>>32) > mod)
	{
		q[s1.mq-1]--; /* adjust1 */
		goto subtract;
	}
	if((uint32_t)(proizv>>32) < mod)
		goto subtract;
	if((uint32_t)proizv > u[s1.mu-2])
		q[s1.mq-1]--; /* adjust1 */
subtract:
	t = s1.mv;
	prom = 0;
	prev = 0;
	for(i = 0; i < (int)(s1.mv); i++)
	{
		prom = (uint64_t)(div_) * v[i]; /* Mnozhim chastnoe na razrjad delitelja */
		prom += (uint64_t)prev;
		up = u[s1.mu - t + i]; /* Dalee otnimaem ot delimogo */
		up = up - (uint32_t)prom;
		u[s1.mu - t + i] = (uint32_t)up;
		if(up>>32 == ASYM_MINUS_1) {/* Nado uchest' zajom */
			prom += ASYM_ELDER;
			prom += ASYM_ELDER;
		}
		prev = (uint32_t) (prom >> 32);
	}
	prom = (uint64_t)(u[s1.mu - t + s1.mv]) - prev;
	u[s1.mu - t + s1.mv] = (uint32_t)prom;
	if((prom >> 32) == ASYM_MINUS_1) {
	/* rezul'tat vychitanija otricatel'nyj! nado pribavit' delitel'
	 * I ne zabyt' umen'shit' znachenie chastnogo na 1. 
	 */
		q[s1.mq - 1] --;
		prom = 0;
		for (i = 0; i < (int)(s1.mv); i++)
		{
			prom += v[i];
			prom += u[s1.mu-t+i];
			u[s1.mu - t + i] = (uint32_t)prom;
			prom = prom >> 32;
		}
		u[s1.mu - t + s1.mv] += (uint32_t)prom;
	
	}
	s1.end_u -= 4;
	s1.end_q -= 4;
	s1.mq --;
	s1.mu --;
	if(s1.mq > 0)
		goto mainLoop;
	s1.mu = rez_mu;
	s1.mq = rez_mq;
	if(s1.sd >0 ) /* Denormalizing */
	{
		p1 = 0;
		p2 = 0;
		for (i = 0; i < (int)(s1.mu - 1); i++)
		{
			p1 = u[i];
			p1 = p1>>s1.sd;
			p2 = (uint64_t)u[i+1]<<32;
			p2 = p2>>s1.sd;
			p1 ^= p2;
			u[i] = (uint32_t)p1;
		}	
		u[s1.mu-1] = u[s1.mu-1]>>s1.sd; /* Last digit... */
		/* Shift divider */
		for(i = 0; i < (int)(s1.mv - 1); i++)
		{
			p1 = v[i];
			p1 = p1>>s1.sd;
			p2 = (uint64_t)v[i+1]<<32;
			p2 = p2>>s1.sd;
			p1 ^= p2;
			v[i] = (uint32_t)p1;
		}
		v[s1.mv-1] = v[s1.mv-1]>>s1.sd; /* Last digit...*/
	}
	rc = 1;
end:

	SZZS(addr_r_u);
	SZZS(sd);
	SZZS(rez_mu);
	SZZS(rez_mq);
	SZZS(delitelr);
	SZZS(div_);
	SZZS(mod);
	SZZS(prev);
	SZZS(p1);
	SZZS(p2);
	SZZS(delimoer);
	SZZS(proizv);
	SZZS(prom);
	SZZS(up);
	SZZS(s1);
	return rc;
}

/*********************************************************************************************	*/
/*Целочисленное деление: q = u/v, u = u%v, k1- разрядность делимого. Разрядность делителя: 16 	*/
/*В параметре s1 сохраняются некоторые промежуточные вычисления во время операции. Не должен быть NULL*/
/*(!)Функция возвращает 0 в случае ошибки; 1 в случае успешного завершения операции; 2 - если делитель оказался больше делимого,
в этом случае содержимое делимого является остатком от деления			*/
int gostc_Divide_c16(uint32_t *u,uint32_t *v, uint32_t *q, uint32_t k1)
{
	uint32_t addr_r_u=0, sd=0, rez_mu=0, rez_mq=0, delitelr=0, div_=0, mod=0;
	uint32_t prev=0;
	uint64_t p1=0, p2=0, delimoer=0, proizv=0, prom=0, up=0;
	int i=0, t=0;
	int rc = 0;
	gostc_BUF_ECP s1={0};

	if((u == NULL)||(v == NULL)||(q == NULL))
		return 0;

	addr_r_u = k1;
/*определим первый значимый разряд.*/
	while(addr_r_u && !u[addr_r_u - 1])
	{	
		addr_r_u--;
	}
	/*поймали.*/
	s1.mu = addr_r_u;/*Сохранили разрядность делимого*/
	
	addr_r_u = 16;
	while(addr_r_u && !v[addr_r_u - 1])
	{
		addr_r_u--;
	}
	s1.mv = addr_r_u;
	

	/*Разобрались с делителем
	////Определим теперь кол-во незначащих старших двоичных разрядов делителя*/
	sd = v[s1.mv - 1];
	i = 31;
	sd = 1 << 31;
	while (!(v[s1.mv - 1]&sd))
	{
		sd = sd>>1;
		i--;
	}
	sd = 31 - i;
	s1.sd = sd;/*Сохранили.*/
	
	/* делитель по разрядам больше делимого.*/
	if(s1.mv > s1.mu)
	{
		rc = 2;/*warning*/
		goto end;
	}
	
	s1.mq = s1.mu - s1.mv + 1;/*Сохранили mq*/
	/*Сохранили старший разряд MQ*/
	/*Нормализация*/
	if(s1.sd!=0)
	{

		p1 = 0,p2 = 0;
		for(i=0;(uint32_t)i<s1.mu-1;i++)
		{
			p1 = u[s1.mu -1 -i];
			p1 = p1<<s1.sd;
			p2 = u[s1.mu -2 -i];
			p2 = p2<<s1.sd;
			p2 = p2>>32;
			p1^=p2;
			u[s1.mu-1-i] = (uint32_t)p1;;
	
		}	
		u[0] = u[0]<<s1.sd;/*Последний разряд...*/

	
		/*Сдвигаем делитель*/
	
		for(i=0;(uint32_t)i<s1.mv-1;i++)
		{
			p1 = v[s1.mv -1 -i];
			p1 = p1<<s1.sd;
			p2 = v[s1.mv -2 -i];
			p2 = p2<<s1.sd;
			p2 = p2>>32;
			p1^=p2;
			v[s1.mv-1-i] = (uint32_t)p1;
		}
		v[0] = v[0]<<s1.sd;/*Последний разряд...*/
		
	}

	rez_mu = s1.mu;
	rez_mq = s1.mq;
mainLoop64:

	delimoer = u[s1.mu];
	delitelr = v[s1.mv-1];
	div_ = 0;

	if((uint32_t)delimoer < delitelr)
		goto divide64;
	/*Сохраним ASYM_MINUS_1 в старшем разряде частного*/
	q[s1.mq -1] = ASYM_MINUS_1;
	goto subtract64;
divide64:
	delimoer = (delimoer << 32) + (uint32_t)u[s1.mu -1];

	div_ = (uint32_t)(delimoer/delitelr);/*Частное */
	mod = delimoer%delitelr;/*Остаток*/

	s1.remainder = mod;/*сохраним их*/
	q[s1.mq -1] = div_;

	proizv = (uint64_t)(div_) * v[s1.mv-2];/*Множим предполагаемое частное на следующий разряд делителя.*/
	
	if((uint32_t)(proizv>>32) > mod)
	{
		q[s1.mq-1]--;/*adjuct1*/
		div_--;/*Вот где хрень водилась!!!*/
		goto subtract64;
	}
	if((uint32_t)(proizv>>32) < mod)
		goto subtract64;
	if((uint32_t)proizv > u[s1.mu-2])
		q[s1.mq-1]--;/*adjust1*/
subtract64:
	t = s1.mv;
	prom = 0;
	prev = 0;
	for( i=0;(uint32_t)i<s1.mv;i++)
	{
		prom = (uint64_t)(div_) * v[i];/*Множим частное на разряд делителя*/
		prom += (uint64_t)prev;

		up = u[s1.mu - t + i];/*Далее отнимаем от делимого*/
		up = up - (uint32_t)prom;

		u[s1.mu - t + i] = (uint32_t)up;
		if(up>>32 == ASYM_MINUS_1)/*Надо учесть заём*/
			prom += 0x100000000;
		prev = (uint32_t) (prom >> 32);
	}
	prom = (uint64_t)(u[s1.mu - t + s1.mv]) - prev;
	u[s1.mu - t + s1.mv] = (uint32_t)prom;
	if((prom >> 32) == ASYM_MINUS_1)/*результат вычитания отрицательный! надо прибавить делитель*/
	{	/*И не забыть уменьшить значение частного на 1.*/
		q[s1.mq - 1] --;
		prom = 0;
		for(i=0;(uint32_t)i<s1.mv;i++)
		{
			prom += v[i];
			prom += u[s1.mu-t+i];
			u[s1.mu - t + i] = (uint32_t)prom;
			prom = prom >> 32;
		}
		u[s1.mu - t + s1.mv] += (uint32_t)prom;

	}
	s1.end_u -= 4;
	s1.end_q -= 4;
	s1.mq --;
	s1.mu --;
	if(s1.mq > 0)
		goto mainLoop64;
	s1.mu = rez_mu;
	s1.mq = rez_mq;
	if(s1.sd >0 )/*Денормализация*/
	{
		p1 = 0;
		p2 = 0;
		for(i=0;(uint32_t)i<s1.mu-1;i++)
		{
			p1 = u[i];
			p1 = p1>>s1.sd;
			p2 = (uint64_t)u[i+1]<<32;
			p2 = p2>>s1.sd;
			p1^=p2;
			u[i] = (uint32_t)p1;
		}	
		u[s1.mu-1] = u[s1.mu-1]>>s1.sd;/*Последний разряд*/
		/*Сдвигаем делитель*/
		for(i=0;(uint32_t)i<s1.mv-1;i++)
		{
			p1 = v[i];
			p1 = p1>>s1.sd;
			p2 = (uint64_t)v[i+1]<<32;
			p2 = p2>>s1.sd;
			p1^=p2;
			v[i] = (uint32_t)p1;
		}
		v[s1.mv-1] = v[s1.mv-1]>>s1.sd;/*Последний разряд*/
			
	}

	rc = 1;
end:
	SZZS(addr_r_u);
	SZZS(sd);
	SZZS(rez_mu);
	SZZS(rez_mq);
	SZZS(delitelr);
	SZZS(div_);
	SZZS(mod);
	SZZS(prev);
	SZZS(p1);
	SZZS(p2);
	SZZS(delimoer);
	SZZS(proizv);
	SZZS(prom);
	SZZS(up);
	SZZS(i);
	SZZS(t);
	SZZS(s1);
	return rc;
}


/* Compare:
 * B > A - returns 0
 * B < A - returns 1
 * B == A - returns 2
-1 - ERROR
 */

int gostc_Comp_C(const uint32_t* A, const uint32_t* B, const uint32_t l)
{
	uint32_t i=0;
	if(!A || !B)/*Params error!*/
		return -1;

	for(i=0;i<l;i++)
	{
		if(B[l - i - 1] > A[l - i - 1])
			return 0;
		if(B[l - i - 1] < A[l - i - 1])
			return 1;
	}
	return 2;
}

/***************************************************************************/
/*Умножение двух чисел: 8-разрядное b и l-разрядное a. Результат пишем в c.*/
int gostc_Mul(const uint32_t *a, const uint32_t *b,uint32_t *c,uint32_t l)
{
	uint32_t i=0, j=0;
	uint64_t p1=0, p2=0;
	
	if(!a || !b || !c || l==0|| l > 32)
		return -1;/*Parameters error!*/

	for(i=0;i<l*2+1;i++)
		c[i] = 0; /* занулим с */
	
	for(j=0;j<l;j++)
	{
		p1 = 0;
		p2 = 0;
		for(i=0;i<ASYM_DEF_SIZE>>2;i++) {
			p1 = b[i]; /* Берем i- й разряд В */
			p1 = p1 * (uint64_t)a[j]; /* Множим его на j-й разряд a */
			p1 = p1 + p2; /* Складываем со старшей частью произведения от прошлой итерации */
			p2 = (uint64_t)c[i+j];
			p2 = p2 + p1;
			c[i+j] = (uint32_t)p2;
			p2 = p2 >> 32;
		}
		c[(ASYM_DEF_SIZE>>2)+j] = (uint32_t)p2;
	}
	SZZS(p1);
	SZZS(p2);
	return 0; /*OK*/
}

/******************************************************************************/
/*Умножение двух чисел: 16-разрядное b и l-разрядное a. Результат пишем в c.  */
int gostc_Mul16(const uint32_t *a, const uint32_t *b,uint32_t *c,uint32_t l)
{
    	int i=0;int j=0;
   	uint64_t p1 =0; uint64_t p2 = 0;
	if(!a || !b || !c || l==0|| l > ASYM_DEF_512_SIZE)
		return -1;/*Ошибка передачи параметров*/   
 
	for(i=0;(uint32_t)i<l*2+1;i++)
		c[i] = 0;
	
	for(j=0;(uint32_t)j<l;j++)
	{
		p1 = 0,p2 = 0;
		for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
		{
			p1 = b[i];/*Берем i- й разряд В*/
			p1 = p1 * (uint64_t)a[j]; /*Множим его на j-й разряд a */
			p1 = p1 + p2; /*Складываем со старшей частью произведения от прошлой итерации*/
			p2 = (uint64_t)c[i+j];
			p2 = p2 + p1;
			c[i+j] = (uint32_t)p2;
			p2 = p2 >> 32;
		}
		c[(ASYM_DEF_512_SIZE>>2)+j] = (uint32_t)p2;
	}
	SZZS(p1);
	SZZS(p2);
	return 0;/*OK*/
}
/*************************************************************************************/
/*Умножение 8-разрядного числа a на 2 c приведением по модулю p. Результат пишем в n.*/
int gostc_Mul_2C(uint32_t *a, const uint32_t *p, uint32_t *n)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	uint32_t c[ASYM_DEF_SIZE>>2]={0};
	int i=0, l=0,rc = -1;
	if(!a || !p || !n)/*Ошибка передачи параметров!*/
		return rc;
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)a[i];
		p1 = p1 << 1; /* Множим на два */
		p1 = p1 + p2; /* Складываем с переносом от предыдущей итерации */
		c[i] = (uint32_t)p1;
		p2 = p1 >> 32; /* Запрминаем перенос (или его отсутствие) */
	}
	if(p2 == 1)/* Перенос был - значит вычитание */
	{
		rc = gostc_Sub_C((uint32_t*)c,(uint32_t*)p,(uint32_t*)n);
		goto end;
	}
	l = gostc_Comp_C((uint32_t*)c,(uint32_t*)p,ASYM_DEF_SIZE>>2);
	if(l<0)
		goto end;/*Ошибка*/
	if(l == 1)
	{
		rc = gostc_Sub_C((uint32_t*)c,(uint32_t*)p,(uint32_t*)n);
		goto end;
	}
	else
	{
		memcpy(n,c,ASYM_DEF_SIZE);
	}
	rc  = 0; /*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZ(c);
	return rc;
}
/**************************************************************************************/
/*Умножение 16-разрядного числа a на 2 c приведением по модулю p. Результат пишем в n.*/
int gostc_Mul16_2C(uint32_t *a,const uint32_t *p, uint32_t *n)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	uint32_t c[ASYM_DEF_512_SIZE>>2]={0};
	int i=0; int l = 0,rc=-1;
	if(!a ||  !p || !n)/*Ошибка передачи параметров!*/
		return rc;

	for(i=0;i<(ASYM_DEF_512_SIZE>>2);i++)
	{
		p1 = (uint64_t)a[i];
		p1 = p1 << 1;/*Множим на два*/
		p1 = p1 + p2; /*Складываем с переносом от предыдущей итерации*/
		c[i] = (uint32_t)p1;
		p2 = p1 >> 32; /*Запрминаем перенос (или его отсутствие)*/
	}
	if(p2 == 1)/*Перенос был - значит вычитание*/
	{
		rc = gostc_Sub_C64(c,p,n);
		goto end;
	}
	l = gostc_Comp_C(c,p,ASYM_DEF_512_SIZE>>2);
	if(l<0)
		goto end;/*Ошибка*/
	if(l == 1)
	{
		rc = gostc_Sub_C64(c,p,n);
		goto end;
	}
	else
	{
		memcpy(n,c,ASYM_DEF_512_SIZE);
	}
	rc = 0;/*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZ(c);
	return rc;
}

/*************************************************************************************/
/*Умножение 8-разрядного числа a на 3 c приведением по модулю p. Результат пишем в n.*/

int gostc_Mul_3_Mod(uint32_t *a,const uint32_t *p,uint32_t *n)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	uint32_t old=0,c[ASYM_DEF_SIZE>>2]={0};
	int i=0,rc=-1;
	int l=0;

	if(!a || !p || !n)/*Ошибка передачи параметров!*/
		return rc;

	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)a[i];/* Берем разряд */
		p1 = p1 * 3 ;/* Множим на три */
		p1 = p1 + p2; /* Складываем со старшей частью от предыдущей итерации */
		c[i] = (uint32_t)p1;
		p2 = p1 >> 32; /* Запоминаем старшую часть (или её отсутствие) */
	}	
	old = (uint32_t)p2;/* Сохраняем старшую часть в 9-м разряде */
	p2 = 0;
	l = 0;
	/* Проверяем старший разряд */
	if(old > 0) /* вычитаем 1-й раз */
	{
		goto Subtr1;
	}
	l = gostc_Comp_C((uint32_t*)c,p,ASYM_DEF_SIZE>>2);
	if(l<0)
	{
		rc =- 2;/*ОШибка*/
		goto end;
	}	
	if(l == 1)
		goto Subtr1;
	else
		goto g3;
Subtr1:							/* Процедура вычитания */
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)c[i];
		p1 = p1 - (uint64_t)p[i];/* Отнимаем разряд B */
		p1 = p1 - p2; /* Учитываем заём с пред. разряда */
		c[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}
	if((p2)&&(old > 0))
		old --;
	p2 = 0;
g3:
	if(old > 0) /* вычитаем 2-й раз */
	{
		goto Subtr2;
	}
	l = gostc_Comp_C((uint32_t*)c,p,ASYM_DEF_SIZE>>2);
	if(l<0)
	{
		rc =- 2;/*ОШибка*/
		goto end;
	}	
	if(l == 1)
		goto Subtr2;
	else
		goto EndOp;
Subtr2:							/* Процедура вычитания (2 раз) */
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)c[i];
		p1 = p1 - (uint64_t)p[i];/* Отнимаем разряд B */
		p1 = p1 - p2; /* Учитываем заём с пред. разряда */
		c[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}			
EndOp:
	memcpy(n,c,ASYM_DEF_SIZE);
	rc = 0;/*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZ(c);
	return rc;
}


/*************************************************************************************/
/*Деление 8-разрядного числа a на 2 c приведением по модулю p. Результат пишем в rez.*/
int gostc_div2_C(uint32_t *a,const uint32_t *p2,uint32_t *rez)
{
	uint64_t p1 = 0;
	uint32_t md = 0;
	uint32_t c[ASYM_DEF_SIZE>>2]={0};
	int i=0,rc=-1;
	if(!a || !p2 || !rez)
		return rc;/*Ошибка передачи параметров*/

	md = a[0]&1;
	memcpy(c,a,ASYM_DEF_SIZE);
	for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
	{
		c[i] = c[i]>>1;
		if((c[i+1]&1) == 1)
			c[i] = c[i] + ASYM_ELDER;
	}
	c[(ASYM_DEF_SIZE>>2)-1] = c[(ASYM_DEF_SIZE>>2)-1] >> 1;
	if(!md) /* По модулю приводить не надо - копировать результат и возврат. */
	{
		memcpy(rez,c,ASYM_DEF_SIZE);
		rc = 0;/*OK*/
		goto end;
	}
	/* Приведение по модулю */
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = p1 + (uint64_t)c[i];
		p1 = p1 + (uint64_t)p2[i];
		rez[i] = (uint32_t)p1;
		p1 = p1 >> 32;
	}
	rc =0;/*OK*/
end:
	SZZS(p1);
	SZZS(md);
	SZZ(c);
	return rc;
}

/*************************************************************************************/
/*Деление 16-разрядного числа a на 2 c приведением по модулю p. Результат пишем в rez.*/
int gostc_div2_64(uint32_t *a,const uint32_t *p2,uint32_t *rez)
{
	uint64_t p1 = 0;
	int i=0, rc = -1;
	uint32_t md=0; 
	uint32_t c[ASYM_DEF_512_SIZE>>2]={0};
	
	if(!a ||  !p2 || !rez)
		return rc;/*Ошибка передачи параметров*/
	
	md = a[0]&1;

	memcpy(c,a,ASYM_DEF_512_SIZE);
	for( i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
	{
		c[i] = c[i]>>1;
		if((c[i+1]&1) == 1)
			c[i] = c[i] + ASYM_ELDER;
	}
	c[(ASYM_DEF_512_SIZE>>2)-1] = c[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
	if(!md)/*По модулю приводить не надо - копировать результат и возврат.*/
	{
		memcpy(rez,c,ASYM_DEF_512_SIZE);
		rc =0;		/*OK*/
		goto end;
	}
	/*Приведение по модулю*/
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 = p1 + (uint64_t)c[i];
		p1 = p1 + (uint64_t)p2[i];
		rez[i] = (uint32_t)p1;
		p1 = p1 >> 32;
	}
	rc=0;/*OK*/
end:
	SZZS(p1);
	SZZS(md);
	return rc;
}



/*Приведение (c+8)-разрядного числа a по 8-разрядному модулю p в арифметике Монтгомери. e - значение EpsilonQ,
полученное с помощью gostc_EpsilonQ
Буфер числа a должен содержать один дополнительный разряд на случай переноса, т.о. размер буфера a д.б. (с+8)+1.
Результатом правильного приведения по модулю будет число, помещенное в буфер a начиная с 8-го разряда. 
разряды а с 0 по 7 д.б. равны 0.
*/
int gostc_ModP_C(uint32_t *a,const uint32_t* p, const uint32_t e, uint32_t c)
{
	uint64_t p1=0;
	uint64_t p2=0;
	uint64_t X0_on_E = 0;
	uint32_t i=0;
	uint32_t j=0;
	int r=0,rc=-1;
	uint32_t pr[ASYM_DEF_SIZE>>2]={0};
	
	if(!a || !p || e==0 || c==0)
		return rc;/*Ошибка передачи параметров*/

	for(i = 0;i<c;i++)
	{
		X0_on_E = (uint64_t)a[i] * (uint64_t)e; /* Получили X0 * e
		* Далее получившееся произв-е в цикле множим на p */
		j = 0;
		for(j=0;j<ASYM_DEF_SIZE>>2;j++)
		{
			p1 = (uint64_t)p[j] * (uint32_t)X0_on_E;
			p1 = p1 + p2;/* Складываем со старшей частью произведения от предыдущей итерации */
			p2 = (uint64_t)a[i+j] + (uint32_t)p1;
			a[i+j] = (uint32_t)p2;
			p1 = p1 >> 32;
			p2 = p2 >>32;
			p2 = p1 + p2;/* Сохранили разряд */
		}
		p2 = p2 + (uint64_t)a[i+j];	
		a[i+j] = (uint32_t)p2;
		p2 = p2 >> 32;
		while(p2 == 1) /* пока перенос */
		{
			j++;
			p2 = p2 + (uint64_t)a[i+j];	
			a[i+j] = (uint32_t)p2;
			p2 = p2 >> 32;
		}
	}
	if(a[(ASYM_DEF_SIZE>>2)+c] != 0)
	{
		if(gostc_Sub_C(a+c,p,pr)<0)
		{
			rc=-2;
			goto end;
		}
		memcpy(a+c,pr,ASYM_DEF_SIZE);
		rc =0; /*OK*/
		goto end;
	}
	r = gostc_Comp_C(a+c,p,ASYM_DEF_SIZE>>2);
	if(r<0)
	{
		rc=-2;
		goto end;
		
	}	
	if(r == 1)
	{	
		if(gostc_Sub_C(a+c,p,pr)<0)
		{
			rc=-2;
			goto end;
		}
		memcpy(a+c,pr,ASYM_DEF_SIZE);
	}
	rc =0;/*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZS(X0_on_E);
	SZZS(i);
	SZZS(j);
	SZZS(r);
	SZZ(pr);
	return rc;
}


/*************************************************************************************/
/*Приведение (c+16)-разрядного числа a по 16-разрядному модулю p в арифметике Монтгомери. e - значение EpsilonQ,
полученное с помощью gostc_EpsilonQ*/
/*Буфер числа a должен содержать один дополнительный разряд на случай переноса, т.о. размер буфера a д.б. (с+16)+1.*/
/*Результатом правильного приведения по модулю будет число, помещенное в буфер a начиная с 16-го разряда. 
разряды а с 0 по 15 д.б. равны 0.*/
int gostc_ModP16(uint32_t *a,const uint32_t* p, const uint32_t e, uint32_t c)
{
	uint64_t p1=0;
	uint64_t p2=0;
	uint64_t X0_on_E = 0;
	int i=0;
	int j=0;
	int r = 0,rc=-1;
	uint32_t pr[ASYM_DEF_512_SIZE>>2]={0};

	if(!a || !p || e==0 || c==0)
		return rc;/*Ошибка передачи параметров*/

	for(i = 0;(uint32_t)i<c;i++)
	{
		X0_on_E = (uint64_t)a[i] * (uint64_t)e; /*Получили X0 * e*/
		/*Далее получившееся произв-е в цикле множим на p*/
		j = 0;
		for(j=0;j<ASYM_DEF_512_SIZE>>2;j++)
		{
			p1 = (uint64_t)p[j] * (uint32_t)X0_on_E;
			p1 = p1 + p2;/*Складываем со старшей частью произведения от предыдущей итерации*/
			p2 = (uint64_t)a[i+j] + (uint32_t)p1;
			a[i+j] = (uint32_t)p2;
			p1 = p1 >> 32;
			p2 = p2 >>32;
			p2 = p1 + p2;/*Сохранили разряд*/
		}
		p2 = p2 + (uint64_t)a[i+j];	
		a[i+j] = (uint32_t)p2;
		p2 = p2 >> 32;
		while(p2 == 1) /*пока перенос*/
		{
			j++;
			p2 = p2 + (uint64_t)a[i+j];	
			a[i+j] = (uint32_t)p2;
			p2 = p2 >> 32;
		}
	}
	if(a[c+(ASYM_DEF_512_SIZE>>2)] != 0)
	{
		if(gostc_Sub_C64(a+c,p,pr)<0)
		{
			rc =- 2;
			goto end;
		}
		memcpy(a+c,pr,ASYM_DEF_512_SIZE);
		rc =0;/*OK*/
		goto end;
	}
	r = gostc_Comp_C(a+c,p,ASYM_DEF_512_SIZE>>2);
	if(r <0)
	{
		rc =- 2;
		goto end;
	}
		
	if(r == 1)
	{
		if(gostc_Sub_C64(a+c,p,pr)<0)
		{
			rc =- 2;
			goto end;
		}
		memcpy(a+c,pr,ASYM_DEF_512_SIZE);
	}
	rc =0; /*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZS(X0_on_E);
	SZZS(i);
	SZZS(j);
	SZZS(r);
	SZZ(pr);
	return rc;
}
