#include <string.h>
#include "gostc_mem.h"

/*############################ OPENSSL_ANCHOR ##############################*/
/*                         DO NOT REMOVE OR EDIT!                           */

#define GCM_MALLOC	malloc
#define GCM_FREE	free
#define GCM_REALLOC	realloc

/*########################## OPENSSL_ANCHOR_END ############################*/

static unsigned char gostc_cleanse_ctr = 0;
/****************************************************************************/
void gostc_cleanse(void *ptr, size_t len)
{
	unsigned char *p = ptr;
	size_t loop = len, ctr = gostc_cleanse_ctr;
	
	while(loop--) {
		*(p++) = (unsigned char)ctr;
		ctr += (17 + ((size_t)p & 0xF));
	}
	p = memchr(ptr, (unsigned char)ctr, len);
	if(p) ctr += (63 + (size_t)p);
	gostc_cleanse_ctr = (unsigned char)ctr;
}
/****************************************************************************/
void *gostc_calloc(size_t nmemb, size_t size)
{
	size_t sz;

	sz = nmemb * size;
	return gostc_malloc(sz);
}
/****************************************************************************/
void *gostc_malloc(size_t size)
{
	uint32_t sz;
	uint8_t* p;

	if (!size) return NULL;
	sz = size;
	p = GCM_MALLOC(size + 4);
	bzero(p, size + 4);
	/* little-endian */
	p[0] = sz;
	p[1] = sz >> 8;
	p[2] = sz >> 16;
	p[3] = sz >> 24;

	return p + 4;
}
/****************************************************************************/
void gostc_free(void *ptr)
{
	uint8_t* p = ptr;
	uint32_t sz;

	if (!p) return;
	p -= 4;
	sz = p[3];  sz <<= 8; 
	sz |= p[2]; sz <<= 8;
	sz |= p[1]; sz <<= 8;
	sz |= p[0];
	gostc_cleanse(p + 4, sz);
	GCM_FREE(p);
}
/****************************************************************************/
void *gostc_realloc(void *ptr, size_t size)
{
	uint32_t osz, sz;
	uint8_t* p = ptr;
	uint8_t* p2 = NULL;

	if (!ptr) return gostc_malloc(size);
	if (!size) { gostc_free(ptr); return NULL; }

	p -= 4;
	osz = p[3]; osz <<= 8; 
	osz |= p[2];osz <<= 8;
	osz |= p[1];osz <<= 8;
	osz |= p[0];

	p2 = gostc_malloc(size);

	if (size < osz) sz = size; else sz = osz;
	memcpy(p2, ptr, sz);

	gostc_free(ptr);

	return p2;
}
/****************************************************************************/

