#include "gostc_mem.h"
#include "gostc_pkcs5.h"
#include "gostc_hash.h"

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))



static int Concatenate(void *buf1,uint16_t sz1, void *buf2,uint16_t sz2,
	void *buf3, uint16_t *sz3)
{
	if(!buf1 || !buf2 || !buf3)
		return 0;

	memcpy((uint8_t*)buf3,buf1,sz1);
	memcpy((uint8_t*)buf3+sz1,buf2,sz2);
	*sz3 = sz1 + sz2;
	return 1;
}

/* ///// U_i = PRF(P, U_i-1) */
static int PRF_Compute(void *Key, uint16_t szKey, void*D, uint16_t szd, void *res)
{
	uint8_t *k = NULL;
	int i = 0;
	uint16_t sz = 0;
	uint16_t szprom = 0;
	uint32_t h1[8];
	uint32_t h2[8];
	void* h = NULL;
	uint8_t *bufprom1 = NULL;
	int rc = 0;

	if (!Key) return 0;
	ZZ(h1);
	ZZ(h2);

	k = gostc_malloc(szKey);
	if(!k) return 0;
	memcpy(k, Key, szKey);

	/* ////P[i]^0x36; */
	for(i = 0; i < szKey; i++) k[i] ^= 0x36;
	
	szprom = szKey + szd;
	bufprom1 = gostc_malloc(szprom);
	if(!bufprom1) {
		rc = 0;
		goto end;
	}
	bzero(bufprom1, szprom);

	/* /// Concat Key[i]^0x36 and D */
	Concatenate(k, szKey, D, szd, bufprom1, &sz);
	
	/* /////Calc h1 = hash(Key[i]^0x36 || D) */
	h = gostc_hash_open();
	if(!h) {
		rc = 0;
		goto end;
	}
	gostc_hash_do(h, bufprom1, sz);
	gostc_hash_close(h, h1);
	h = NULL;

/* 	bzero(bufprom1, szprom); - zeroed by gostc_free */
	gostc_free(bufprom1);
	bufprom1 = NULL;
	
	/* ////// Calc Key[i]^0x5c */
	szprom = szKey + 32 + 1;
	bufprom1 = gostc_malloc(szprom);
	if(!bufprom1) {
		rc = 0;
		goto end;
	}
	bzero(bufprom1, szprom);
	
	memcpy(k, Key, szKey);
	for(i = 0; i < szKey; i++)
		k[i] ^= 0x5c;

	/* ////////Key[i]^0x5c || h1 */
	Concatenate(k, szKey, h1, 32, bufprom1, &sz);
	
	h = gostc_hash_open();
	if(!h) {
		rc = 0;
		goto end;
	}
	gostc_hash_do(h, bufprom1, sz);
	gostc_hash_close(h, h2);
	h = NULL;
	
	memcpy(res, h2, 32);
	rc = 1;

end:
	if (k) {
/*		bzero(k, szKey); - zeroed by gostc_free */
		gostc_free(k);
	}
	if (bufprom1) {
/*		bzero(bufprom1, szprom); - zeroed by gostc_free */
		gostc_free(bufprom1);
	}
	SZZ(h1);
	SZZ(h2);
	return rc;
}

/* ///////////// In RFC  F = U_1^U_2^...^U_c */
static int Compute_F(void *P,uint16_t sizeP, void*S, uint16_t sizeS, 
	uint32_t c, uint32_t i,void *F)
{
/*	uint32_t Zero[8] = {0,0,0,0,0,0,0,0}; */
	uint32_t U[8],Unext[8],res[8];
	uint8_t *sprom = NULL;
	uint16_t j = 0, k = 0;
	int rc = 0;

	sprom = gostc_malloc(sizeS + 4);

	memcpy(sprom, S, sizeS);
	memcpy(sprom + sizeS, &i, 4); /* sprom = S || i */
	/* calc U_1 = PDF(P, S||i) */
	if(!PRF_Compute(P, sizeP, sprom, sizeS + 4, Unext)) {
		rc = 0;
		goto end;
	}
	memcpy(res, Unext, 32);
	/* Calc U_i = PDF(P, U_i-1) */
	/* res ^= U_i */

	for(j = 1; j < c; j++) {
		memcpy(U, Unext, 32);
		if(!PRF_Compute(P, sizeP, U, 32, Unext))
		{
			rc = 0;
			goto end;
		}
		for(k=0;k<8;k++)
			res[k] ^= Unext[k];
	}
	/* The result */
	memcpy(F, res, 32);
	rc = 1;
end:
	if (sprom) {
/*		bzero(sprom, sizeS + 4); - zeroed by gostc_free */
		gostc_free(sprom);
	}
	SZZ(res);
	SZZ(U);
	SZZ(Unext);
	return rc;
}

/* Full algorithm (output key size - 32 bytes, hash size - 32 bytes) */
int gostc_GetPBKDF2(void *passw, uint16_t passlen, void* salt, uint16_t saltlen,
	uint16_t iternum,void *ResKey)
{
	int keylen_div_hlen = 1;
	uint8_t key[32];
	void* hh = NULL;
	int rc = 1;

/* VOVA - If password len < 32 - it should be padded with 0s
 * If > 32 - password hash should be computed */

	ZZ(key);
	if (passlen) {
		if (passlen < 33) memcpy(key, passw, passlen);
		else {
			hh = gostc_hash_open();
			gostc_hash_do(hh, passw, passlen);
			gostc_hash_close(hh, key);
		}
	}

	if (!Compute_F(key, 32, salt, saltlen, iternum, keylen_div_hlen, ResKey)) {
		rc = 1;
		goto end;
	}
	rc = 0;
end:
	SZZ(key);
	return rc;
}

/*
int _tmain(int argc, _TCHAR* argv[])
{
	uint32_t P[8] = {0,0,1,1,13,3,3,3};
	uint32_t S[8] = {2,2,4,4,4,4,4,4};
	uint32_t res[8] = {0};
	uint16_t sznew;
	if(!GetPBKDF2(P,1,S,1,3,res))
		printf("Error in getting key\n");
	else
	{
		for(int i=0;i<8;i++)
			printf("0x%8.8X \n",res[i]);
	}
	return 0;
}
*/
/*****************************************************************************/
static int pbkdf2012_F(void* passw, uint16_t passlen, void* salt, uint16_t saltlen, 
	uint32_t c, int n, gostc_hmac12_ctx_t* hctx, uint8_t* tbuf)
{
	int rc = 1;
	uint32_t ci;
	uint8_t inti[4];
	uint8_t u[64];
	int i;

	inti[0] = 0;
	inti[1] = 0;
	inti[2] = n >> 8;
	inti[3] = n;
	if (gostc_hmac12_init(hctx, passw, passlen)) goto end;
	if (gostc_hmac12_do(hctx, salt, saltlen)) goto end;
	if (gostc_hmac12_do(hctx, inti, 4)) goto end;
	if (gostc_hmac12_finish(hctx, tbuf)) goto end;
	memcpy(u, tbuf, 64);

	for (ci = 1; ci < c; ci++) {
		if (gostc_hmac12_init(hctx, passw, passlen)) goto end;
		if (gostc_hmac12_do(hctx, u, 64)) goto end;
		if (gostc_hmac12_finish(hctx, u)) goto end;
		for (i = 0; i < 64; i++) tbuf[i] ^= u[i];
	}
	rc = 0;
end:
	SZZ(u);
	return rc;
}
/*****************************************************************************/
int gostc_GetPBKDF2_2012(void* passw, uint16_t passlen, void* salt, uint16_t saltlen,
	uint32_t c, void* keyout, uint16_t keyoutlen)
{
	int rc = 1;
	uint8_t tbuf[64];
	int n = 1;
	gostc_hmac12_ctx_t hctx;
	int l;
	uint8_t key[64];

	if ((!passw && passlen) || !salt || !saltlen || !c || !keyout || !keyoutlen) 
		return 1;

	/* for optimization */
	if (passlen > 64) {
		if (gostc_hash12_init(&hctx, 0)) goto end;
		if (gostc_hash12_do(&hctx, passw, passlen)) goto end; 
		if (gostc_hash12_finish(&hctx, key)) goto end;
		passlen = 64;
	} else {
		memcpy(key, passw, passlen);
	}

	while (keyoutlen) {
		if (pbkdf2012_F(key, passlen, salt, saltlen, c, n, &hctx, tbuf))
			goto end; 
		l = keyoutlen;
		if (l > 64) l = 64;
		keyoutlen -= l;
		memcpy((uint8_t*)keyout + ((n - 1) << 6), tbuf, l);
		n++;
	}
	rc = 0;
end:
	SZZ(key);
	SZZ(tbuf);
	gostc_hmac12_clear(&hctx);
	return rc;
}
/*****************************************************************************/

