#include "gostc_mem.h"
#include "gostc_sym.h"
#include "gostc_sym_keywrap.h"
#include "gostc_sym_uz.h"

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))/*##Ivan: для структур и переменных*/
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))/*##Ivan: для структур и переменных*/



/****************************************************************************/
int gostc_sym_divers65(void *key_in, void* ukm, int uz, void *key_out)
{
	uint8_t *UKM;
	int i,l;
	uint32_t K[9][8];
	uint32_t S[2];
	uint32_t mask;
	uint8_t zmask[32];
	uint32_t t; 

	if (!key_in || !ukm || !key_out) return 1;
	if (uz < 1 || uz > 5) return 1;
	ZZ(K);
	ZZ(zmask);
	memcpy(&K[0][0], key_in, 32);
	UKM = (uint8_t*)ukm;

	for(i = 0; i < 8; i++) {
		S[0] = S[1] = 0;
		mask = 1;
		for(l = 0; l < 8; l++) {
			if(UKM[i] & mask) {
				S[1] += K[i][l];
			} else {
				S[0] += K[i][l];
			}
			mask <<= 1;
		}

/* OLD FACTOR-TS IMPLEMENTATION *
		t = ntohl(S[0]); S[0] = t;
		t = ntohl(S[1]); S[1] = t;
*/
/* Compatible with OpenSSL */
		t = S[0];
		S[0] = S[1];
		S[1] = t;
/* */
		gostc_sym_EncryptCFB((void*)&K[i][0], (void*)&K[i + 1][0], (void*)S, 
			32, (void*)&K[i][0], (void*)zmask, 
			&gostc_sym_uz[uz][0]);
	}
	memcpy(key_out, &K[8][0], 32);
	SZZ(K);
	SZZ(S);
	SZZS(mask);
	SZZS(t);

	return 0;
}
/****************************************************************************/
int gostc_sym_keywrap(uint8_t* openkey_in, uint8_t* kek_in, uint8_t* ukm, int uz, 
	uint8_t* enckey_out, uint8_t* imit_out)
{
	uint8_t nkek[32];
	int rc = 0;
	gostc_sym_mac_t* mctx = NULL;
	uint8_t zmask[32];
	int i;

	if (!openkey_in || !ukm || !enckey_out || !imit_out || !kek_in) 
		return 1;
	if (uz < 1 || uz > 5) return 1;
	ZZ(zmask);

	/* Diversify KEK */
	if (gostc_sym_divers65(kek_in, ukm, uz, nkek)) { rc = 2; goto end; }

	/* Compute imit */
	mctx = gostc_sym_MacOpen();
	if (!mctx) { rc = 3; goto end; }
	memcpy(&mctx->B, ukm, 4);
	memcpy(&mctx->A, ukm + 4, 4);
	if (gostc_sym_DoMac(mctx, openkey_in, 32, (void*)nkek, (void*)zmask, 
		&gostc_sym_uz[uz][0])) {
		rc = 4;
		goto end;
	}
	if (gostc_sym_MacClose(mctx, imit_out, (void*)nkek, (void*)zmask, 
		&gostc_sym_uz[uz][0])) {
		rc = 5;
		goto end;
	}
	mctx = NULL;

	/* enc CEK in ECB mode */

	for (i = 0; i < 32; i += 8) 
		gostc_sym_ECB((void*)(openkey_in + i), (void*)(enckey_out + i),
			(void*)nkek, (void*)zmask, &gostc_sym_uz[uz][0]);

end:
	SZZ(nkek);
	if (mctx) {
		gostc_free(mctx);
	}
	return rc;
}
/****************************************************************************/
int gostc_sym_keyunwrap(uint8_t* enckey_in, uint8_t* kek_in, uint8_t* imit_in, 
	uint8_t* ukm, int uz, uint8_t* openkey_out)
/* 0-ok, 1-imit failed, >1-error */
{
	int rc = 0;
	gostc_sym_mac_t* mctx = NULL;
	uint8_t zmask[32];
	uint8_t nkek[32];
	uint8_t cek[32];
	int i;
	uint32_t imit;
	uint32_t* pimit;

	if (!enckey_in || !kek_in || !imit_in || !ukm || !openkey_out)
		return 2;
	if (uz < 1 || uz > 5) return 2;
	ZZ(zmask);

	/* Diversify KEK */
	if (gostc_sym_divers65(kek_in, ukm, uz, nkek)) { rc = 2; goto end; }

	/* Decrypt CEK_ENC in ECB */
	for (i = 0; i < 32; i += 8) 
		gostc_sym_ECBBack((void*)(enckey_in + i), (void*)(cek + i),
			(void*)nkek, (void*)zmask, &gostc_sym_uz[uz][0]);

	/* Compute imit */
	mctx = gostc_sym_MacOpen();
	if (!mctx) { rc = 3; goto end; }
	memcpy(&mctx->B, ukm, 4);
	memcpy(&mctx->A, ukm + 4, 4);
	if (gostc_sym_DoMac(mctx, cek, 32, (void*)nkek, (void*)zmask, 
		&gostc_sym_uz[uz][0])) {
		rc = 4;
		goto end;
	}
	if (gostc_sym_MacClose(mctx, &imit, (void*)nkek, (void*)zmask, 
		&gostc_sym_uz[uz][0])) {
		rc = 5;
		goto end;
	}
	mctx = NULL;
	pimit = (uint32_t*)imit_in;
	if (*pimit != imit) { rc = 1; goto end; }

	/* Unwrap ok */
	memcpy(openkey_out, cek, 32);

end:
	if (mctx) {
/*		bzero(mctx, sizeof(gostc_sym_mac_t));  - zeroed by gostc_free */
		gostc_free(mctx);
	}
	SZZ(nkek);
	SZZ(cek);
	SZZS(imit);
	return rc;
}
/****************************************************************************/

