#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "gostc_asym.h"


/*****************************************************************************/
int ReadFile(char* file, uint8_t* buf, int len)
{
	int f = -1;

	if (!f || !buf || !len) return 1;
	f = open(file, O_RDONLY);
	if (f < 0) {
		printf("Failed to open file %s\n", file);
		return 1;
	}
	if (read(f, buf, len) != len) {
		printf("Failed to read file %s. (Maybe wrong length).\n", file);
		close(f);
		return 1;
	}
	close(f);
	return 0;
}
/*****************************************************************************/
int main(int argc, char** argv)
{
	int parset;
	int m512 = 0;
	void *pars;
	uint8_t pubkey[128];
	uint8_t hash[64];
	uint8_t signature[128];
	int rc;


	if (argc != 5) {
		printf("Usage: check_sign <hash_file> <sig_file> <pubkey_file> <par>\n");
		return 2;
	}

	parset = atoi(argv[4]);
	if ((parset < 1) || (parset > 7)) {
		printf("Wrong par\n");
		return 2;
	}
	m512 = parset > 4;
	pars = gostc_asym_LoadParams(parset);
	if (ReadFile(argv[1], hash, m512 ? 64 : 32)) return 1;
	if (ReadFile(argv[2], signature, m512 ? 128: 64)) return 1;
	if (ReadFile(argv[3], pubkey, m512 ? 128 : 64)) return 1;
	rc = gostc_asym_CheckSign(pars, pubkey, hash, signature);
	if (!rc) {
		printf("Signature check NOT passed\n");
		gostc_asym_FreeParams(pars);
		return 1;
	}
	if (rc == 1) {
		printf("Signature check PASSED\n");
		gostc_asym_FreeParams(pars);
		return 0;
	}
	printf("Error: Code %d\n", rc);
	gostc_asym_FreeParams(pars);
	return 2;
}

