#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "gostc_asym.h"

uint8_t PrivKey[64];


/*****************************************************************************/
void mydump(void* ptr,int len)
{
	int i, j;
	unsigned char c, *buf;
	unsigned int cc;

	j = 0;
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j > 15) {
			j = 0;
			printf("\n");
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		j++;
	}
	printf("\n");
}
/*****************************************************************************/
int mycmp(void* v1, void* v2, int cnt) 
{
	int i;
	unsigned char *vc1 = v1;
	unsigned char *vc2 = v2;

	if (!v1 || !v2) return 1;
	for (i = 0; i < cnt; i++) {
		if (vc1[i] != vc2[i]) {
			printf ("Mismatch at byte %d\n", i);
			printf("Calculated:\n");
			mydump(v1, cnt);
			printf("Golden:\n");
			mydump(v2, cnt);
			return 1;
		}
	}
	return 0;
}
/*****************************************************************************/
void myswap(void* dst, void* src, int size)
{
	int i;
	uint8_t *s, *d;

	d = dst;
	s = src;
	if (!s || !d) return;
	for (i = 0; i < size; i++) d[i] = s[size - 1 - i];
}
/*****************************************************************************/
int ReadFile(char* file, uint8_t* buf, int len)
{
	int f = -1;

	if (!f || !buf || !len) return 1;
	f = open(file, O_RDONLY);
	if (f < 0) {
		printf("Failed to open file %s\n", file);
		return 1;
	}
	if (read(f, buf, len) != len) {
		printf("Failed to read file %s\n", file);
		close(f);
		return 1;
	}
	close(f);
	return 0;
}
/*****************************************************************************/
int main(int argc, char** argv)
{
	uint8_t pubkey[128];
	uint8_t privkey[64];
	void *pars;
	int parset;
	int swap = 0;
	int is2012 = 0;

	if (argc != 3 && argc != 4) {
		printf("Usage: mkpub <prkey_file> <par> [swap]\n");
		printf("par = 1-TestGost, 2-CproA(XchA), 3-CproB, 4-CproC(XchB)\n"
			"      5-TC26-512Test, 6-TC26-512A, 7-TC26-512B\n");
		printf("swap - swap private key\n");
		return 1;
	}
	if (argc == 4) swap = 1;
	parset = atoi(argv[2]);
	if ((parset < 1) || (parset > 7)) {
		printf("Wrong par\n");
		return 1;
	}
	is2012 = parset > 4;
	pars = gostc_asym_LoadParams(parset);
	if (ReadFile(argv[1], PrivKey, is2012 ? 64 : 32)) return 1;
	if (swap) {
		printf("PrivKey swapped\n");
		myswap(privkey, PrivKey, is2012 ? 64 : 32);
	} else 
		memcpy(privkey, PrivKey, is2012 ? 64 : 32);

	gostc_asym_GenPublicKey(pars, privkey, pubkey);
	mydump(pubkey, is2012 ? 128 : 64);
	return 0;
}
