#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "gostc_asym.h"



/*****************************************************************************/
void mydump(void* ptr,int len)
{
	int i, j;
	unsigned char c, *buf;
	unsigned int cc;

	j = 0;
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j > 15) {
			j = 0;
			printf("\n");
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		j++;
	}
	printf("\n");
}
/*****************************************************************************/
void myswap(void* dst, void* src, int size)
{
	int i;
	uint8_t *s, *d;

	d = dst;
	s = src;
	if (!s || !d) return;
	for (i = 0; i < size; i++) d[i] = s[size - 1 - i];
}
/*****************************************************************************/
int ReadFile(char* file, uint8_t* buf, int len)
{
	int f = -1;

	if (!f || !buf || !len) return 1;
	f = open(file, O_RDONLY);
	if (f < 0) {
		printf("Failed to open file %s\n", file);
		return 1;
	}
	if (read(f, buf, len) != len) {
		printf("Failed to read file %s. (Maybe wrong length).\n", file);
		close(f);
		return 1;
	}
	close(f);
	return 0;
}
/*****************************************************************************/
int main(int argc, char** argv)
{
	void *pars;
	int parset;
	int swap = 0;
	int m512 = 0;
	uint8_t privkey[64];
	uint8_t hash[64];
	uint8_t rnd[64];
	uint8_t signature[128];
	uint8_t sig_be[128];
	int rc;

	if (argc != 6 && argc != 5) {
		printf("Usage: sign <hash_file> <prkey_file> <rnd_file> <par> [swap]\n");
		printf("par = 1-TestGost, 2-CproA(XchA), 3-CproB, 4-CproC(XchB)\n"
			"      5-TC26-512Test, 6-TC26-512A, 7-TC26-512B\n");
		printf("swap - swap signature to Big-Endian\n");
		return 1;
	}
	if (argc == 6) swap = 1;
	parset = atoi(argv[4]);
	if ((parset < 1) || (parset > 7)) {
		printf("Wrong par\n");
		return 1;
	}
	m512 = parset > 4;
	pars = gostc_asym_LoadParams(parset);
	if (ReadFile(argv[1], hash, m512 ? 64 : 32)) return 1;
	if (ReadFile(argv[2], privkey, m512 ? 64 : 32)) return 1;
	if (ReadFile(argv[3], rnd, m512 ? 64 : 32)) return 1;

	rc = gostc_asym_MakeSign(pars, rnd, hash, privkey, signature);
	if (rc != 1) {
		printf("gostc_asym_MakeSign returned error. Code %d.\n", rc);
		return 1;
	}
	if (!swap) {
		printf("Signature (Little-Endian):\n");
		mydump(signature, m512 ? 128 : 64);
	} else {
		printf("Signature (Big-Endian):\n");
		myswap(sig_be, signature, m512 ? 128 : 64);
		mydump(sig_be, m512 ? 128 : 64);
	}

	return 0;
}

