#include "gostc_asym.h"
#include "gostc_asym_vko.h"
#include "gostc_hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

/* Uncomment for random testing*/
/*#define PRESET_TEST
#define VKO_RANDOM_TEST
*/

#define RANDOM_ITERATIONS 1000

/* For verbosity in CryptoPro test*/
 #define CPRO_VERBOSE 


typedef unsigned char u_int8_t;
#define DBG1 printf

typedef struct vko_t vko_t;
struct vko_t {

	char some_unaligned_bytes[3];

	/* My private value */
	u_int8_t x[32];

	/* My public value */
	u_int8_t gx[64];

	 /* Other public value */
	u_int8_t gy[64];

	/* Shared secret */
	u_int8_t gxy[64];

	/* VKO context */
	void* vko_ctx;
};

/* First preset seeds*/
#define SEED1 0x55555555
#define SEED2 0xCCCCCCCC
#define CPROSET	1
vko_t *VKO1 = NULL;
vko_t *VKO2 = NULL;


int	CP1_cpro_set = 2;
uint8_t CP1_x[32] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x79, 0x65, 0x4b, 0x74, 0x69,
	0x6e, 0x49 };
uint8_t CP1_gx[64] = {
	0x68, 0x38, 0x46, 0xaa, 0xc9, 0x20, 0xd1, 0xaf, 0x0b, 0x83, 
	0x5b, 0x8e, 0xa1, 0x69, 0x74, 0x96, 0xd2, 0x06, 0x8e, 0xea, 
	0x55, 0x04, 0xed, 0xcc, 0x47, 0x27, 0x96, 0xe2, 0x0c, 0x8a, 
	0x65, 0x20,
	0x89, 0x57, 0x76, 0xc7, 0x86, 0x68, 0x95, 0xf6, 0xc0, 0x31, 
	0x6d, 0xac, 0x96, 0x54, 0x56, 0x90, 0x61, 0x91, 0x9b, 0x47, 
	0x9d, 0xfa, 0xe5, 0x7a, 0x98, 0x4f, 0xab, 0xb0, 0x10, 0x05, 
	0xc3, 0x37 };
uint8_t CP1_y[32] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x79, 0x65, 0x4b, 0x70, 0x73,
	0x65, 0x52 };
uint8_t CP1_gy[64] = {
	0x0e, 0x8e, 0x57, 0x76, 0x73, 0xd9, 0xd9, 0x2c, 0xff, 0xad, 
	0x30, 0x2b, 0xcf, 0xa2, 0xc1, 0x41, 0x85, 0x72, 0x7e, 0x6f, 
	0x37, 0x76, 0x30, 0xbb, 0xa4, 0xb0, 0x13, 0xb7, 0xd0, 0xe6, 
	0x4d, 0x59,
	0x24, 0xa7, 0x20, 0x40, 0xeb, 0x72, 0x56, 0x7a, 0xba, 0xe9, 
	0x27, 0x00, 0xb7, 0x5a, 0x48, 0x26, 0xc3, 0xeb, 0xb5, 0x8e, 
	0x3e, 0x97, 0x6c, 0xef, 0x7c, 0xaa, 0xa6, 0xe8, 0x08, 0xe7, 
	0x86, 0x33 };
uint8_t CP1_akey[32] = {
	0xcf, 0xce, 0xc6, 0x05, 0x92, 0x51, 0x08, 0x24, 0x40, 0x35, 
	0x63, 0x6e, 0xc3, 0x7c, 0xba, 0x10, 0x98, 0xdf, 0xbf, 0xc7, 
	0xef, 0xd2, 0x2d, 0x5d, 0xc9, 0xfb, 0x22, 0xdf, 0x6f, 0x4d, 
	0x0c, 0xf7 };

void dump(void* ptr,int len);


/***************************************************************************/
void free_vko(vko_t* vko)
{
	if (!vko) return;
	if (vko->vko_ctx) gostc_vko_FreeParams(vko->vko_ctx);
	free(vko);
}
/***************************************************************************/
void swap32(uint8_t* x)
{
	uint8_t b[32];
	int i;

	if (!x) return;
	for (i = 0; i < 32; i++) 
		b[i] = x[31 - i];
	memcpy(x, b, 32);
}
/***************************************************************************/
void swap3232(uint8_t* gx)
{
	swap32(gx);
	swap32(gx + 32);
}
/***************************************************************************/
vko_t *gost_vko_create(int paramset, int ifseed, uint32_t seed,
	u_int8_t* x)
{
	vko_t *this = malloc(sizeof(vko_t));
	int  rc = 0;
	uint8_t b32[32];
	uint32_t r;
	int i;

	if (!this) return NULL;
	bzero(this, sizeof(vko_t));
	
	if (gostc_vko_LoadParams(paramset, &this->vko_ctx)) {
		free_vko(this);
		return NULL;
	}

	if (ifseed) {
		srand(seed);
		bzero(b32, 32);
		for (i = 0; i < 8; i++) {
			r = (uint32_t)rand();
			memcpy(b32 + i * 4, &r, 4);
		}
		rc = gostc_vko_GenSecretKey(this->vko_ctx, b32, this->x);
		if (rc) {
			DBG1("gost_vko_create: VKOGenSecretKey rc=%d\n", rc);
			free_vko(this);
			return NULL;
		}
		swap32(this->x);
	} else {
		if (!x) {
			free_vko(this);
			return NULL;
		}
		memcpy(this->x, x, 32);
	}
#ifdef CPRO_VERBOSE
	printf("Input secret x (cpro - BE):\n");
	dump(&this->x, 32);
#endif
	memcpy(b32, this->x, 32);
	swap32(b32);
#ifdef CPRO_VERBOSE	
	printf("Swapped secret x (back to normal LE):\n");
	dump(b32, 32);
#endif
	rc = gostc_vko_GenOpenKey(this->vko_ctx, b32, this->gx);
	if (rc) {
		DBG1("gost_vko_create: VKOGenOpenKey rc=%d\n", rc);
		free_vko(this);
		return NULL;
	}
#ifdef CPRO_VERBOSE
	printf("Generated gx (normal LE):\n");
	dump(this->gx, 64);
#endif
	swap3232(this->gx);

#ifdef CPRO_VERBOSE
	printf("Swapping gx (32,32) to match CPRO test:\n");
	dump(this->gx, 64);
#endif

	return this;
}
/***************************************************************************/
int set_other_public_value(vko_t *this, void* value)
{
	uint8_t vko[64];
	uint8_t x[32];
	uint8_t gy[64];
	int rc;
	uint8_t ukm[8];

	if (!this) return 1;
	memcpy(this->gy, value, 64);
	memcpy(x, this->x, 32);
	swap32(x);
#ifdef CPRO_VERBOSE
	printf("Calculating shared secret:\n");
	printf("Input our secret x (swapped, normal, LE):\n");
	dump(x, 32);
#endif
	memcpy(gy, this->gy, 64);
	swap3232(gy);
#ifdef CPRO_VERBOSE
	printf("Input foreign gx (swapped back 32,32, normal, LE):\n");
	dump(gy, 64);
#endif
	bzero(ukm, sizeof(ukm));
	ukm[0] = 1;
	rc = gostc_vko_GenCommonKey(this->vko_ctx, x, gy, ukm, vko);
	if (rc) return rc;
/*	swap3232(vko);*/
	memcpy(this->gxy, vko, 64);
#ifdef CPRO_VERBOSE
	printf("Calculated shared secret gxy (normal LE):\n");
	dump(vko, 64);
#endif
	return 0;
}
/***************************************************************************/
int cmp(void* v1, void* v2, int cnt) 
{
	int i;
	unsigned char *vc1 = v1;
	unsigned char *vc2 = v2;

	if (!v1 || !v2) return 1;
	for (i = 0; i < cnt; i++) {
		if (vc1[i] != vc2[i]) {
			printf ("Mismatch at byte %d\n", i);
			return 1;
		}
	}
	return 0;
}
/***************************************************************************/
void dump(void* ptr,int len)
{
	int i, j;
	unsigned char c, *buf;
	unsigned int cc;

	j = 0;
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j > 15) {
			j = 0;
			printf("\n");
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		j++;
	}
	printf("\n");
}
/***************************************************************************/
void fail(int rc)
{
	printf("FAILED! rc=%d\n", rc);
	free_vko(VKO1);
	free_vko(VKO2);
	exit(rc);
}
/***************************************************************************/
void iterate(uint32_t seed1, uint32_t seed2, int cpro_set)

{
	int rc;

	VKO1 = gost_vko_create(cpro_set, 1, seed1, NULL);
	if (!VKO1) fail(101);
	VKO2 = gost_vko_create(cpro_set, 1, seed2, NULL);
	if (!VKO2) fail(102);

	rc = set_other_public_value(VKO1, VKO2->gx);
	if (rc) fail(rc);
	rc = set_other_public_value(VKO2, VKO1->gx);
	if (rc) fail(rc);
	rc = cmp(VKO1->gxy, VKO2->gxy, 64);
	if (rc) {
		DBG1("vko1 gxy:\n");
		dump(VKO1->gxy, 64);
		DBG1("vko2 gxy:\n");
		dump(VKO2->gxy, 64);
		
		DBG1("\nvko1 x:\n");
		dump(VKO1->x, 32);
		DBG1("vko1 gx:\n");
		dump(VKO1->gx, 64);

		DBG1("\nvko2 y:\n");
		dump(VKO2->x, 32);
		DBG1("vko2 gy:\n");
		dump(VKO2->gx, 64);
		
		fail(rc);
	}
	free_vko(VKO1);
	free_vko(VKO2);
	return;
}
/***************************************************************************/
void HashGXY(uint8_t* gxy64, uint8_t* hash32)
{
	void* hh = NULL;

	if (!gxy64 || !hash32) return;
	hh = gostc_hash_open();
	gostc_hash_do(hh, gxy64, 64);
	gostc_hash_close(hh, hash32);
}
/***************************************************************************/
void cpro_test(int cpro_set, uint8_t* x, uint8_t* gx, 
	uint8_t* y, uint8_t* gy, uint8_t* akey)
{
	int rc;
	uint8_t h[32];


#ifdef CPRO_VERBOSE
	printf("\nFirst peer VKO:\n");
#endif
	VKO1 = gost_vko_create(cpro_set, 0, 0, x);
	if (!VKO1) fail(121);
	rc = cmp(VKO1->gx, gx, 64);
	if (rc) {
		printf("cpro_test:vko1 gx mismatch\n");
		printf("vko1 gx:\n");
		dump(VKO1->gx, 64);
		printf("cpro gx:\n");
		dump(gx, 64);
		fail(123);
	}

#ifdef CPRO_VERBOSE
	printf("\nFirst peer gx matched.\n");
	 
	printf("\nSecond peer VKO:\n");
#endif
	VKO2 = gost_vko_create(cpro_set, 0, 0, y);
	if (!VKO2) fail(122);
	rc = cmp(VKO2->gx, gy, 64);
	if (rc) {
		printf("cpro_test:vko2 gx mismatch\n");
		printf("vko2 gx:\n");
		dump(VKO2->gx, 64);
		printf("cpro gy:\n");
		dump(gy, 64);
		fail(124);
	}
#ifdef CPRO_VERBOSE
	printf("\nSecond peer gx matched.\n");

	printf("\nCalculating shared secret for peer 1:\n\n");
#endif
	
	rc = set_other_public_value(VKO1, VKO2->gx);
	if (rc) fail(rc);

#ifdef CPRO_VERBOSE
	printf("\nCalculating shared secret for peer 2:\n");
#endif

	rc = set_other_public_value(VKO2, VKO1->gx);
	if (rc) fail(rc);
	rc = cmp(VKO1->gxy, VKO2->gxy, 64);
	if (rc) {
		printf("cpro_test: gxy mismatch\n");
		printf("VKO1->gxy:\n");
		dump(VKO1->gxy, 64);
		printf("VKO2->gxy:\n");
		dump(VKO2->gxy, 64);
		fail(125);
	}
#ifdef CPRO_VERBOSE
	printf("\nGXYs of peers match.\n");

	printf("\nHashing gxy (normal, LE):\n");
	dump(VKO1->gxy, 64);
#endif
	HashGXY(VKO1->gxy, h);
#ifdef CPRO_VERBOSE
	printf("\nCalculated hash:\n");
	dump(h, 32);
#endif
	swap32(h);
#ifdef CPRO_VERBOSE
	printf("Swapping hash to match CPRO (BE):\n");
	dump(h, 32);
#endif
	rc = cmp(akey, h, 32);
	if (rc) {
		printf("cpro_test: akey mismatch\n");
		printf("cproakey :\n");
		dump(akey, 32);
		printf("computed hash:\n");
		dump(h, 32);
		fail(126);
	}
	printf("CryptoPro test MATCHED.\n");
	free_vko(VKO1);
	free_vko(VKO2);
}
/***************************************************************************/

int main(int argc, char ** argv)
{
#ifdef VKO_RANDOM_TEST
	uint32_t s1, s2;
	int cpro_set;
	int i;
	
#endif

	if (!gostc_asym_TestLib()) {
		printf("Self-test of GOST R 3410-2001 library - FAILED.\n");
		return 1;
	}
	else
	{
		printf("Self-test of GOST R 3410-2001 library - CORRECT!\n");
	}
/*#ifdef GOSTC_ASYM_TEST_SPEED
	printf("Speed-test gost2012-testSet\n");
	if(!gostc_asym_TestLibSpeed(&sp, &sp2,5))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
		return 0;
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
#endif*/
	
#ifdef PRESET_TEST
	printf("Testing preset seeds: 0x%.8X, 0x%.8X, par=%d\n", SEED1, SEED2,
		CPROSET);
	iterate(SEED1, SEED2, CPROSET);
#endif

	printf("CryptoPro test1:\n");
	cpro_test(CP1_cpro_set, CP1_x, CP1_gx, CP1_y, CP1_gy, CP1_akey);

#ifdef VKO_RANDOM_TEST
	printf("Random testing:\n");
	srand(time(NULL));
	for (i = 0; i < RANDOM_ITERATIONS; i++) {
		s1 = rand();
		s2 = rand();
		cpro_set = rand() & 1;
		cpro_set++;
		printf("0x%.8X, 0x%.8X, par=%d\n", s1, s2, cpro_set);
		iterate(s1, s2, cpro_set);
	}
#endif
	return 0;
}
/***************************************************************************/

