


#ifdef WIN32

#include "stdafx.h"
#include <io.h>
#include <windows.h>
#include "pkcs15dll.h"

#ifndef bzero
#define bzero ZeroMemory
#endif 

#else

#include <strings.h>
#include <unistd.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

#endif

#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <sys/types.h>

#include <pkcs15container.h>

#define RND_EXTRA_LEN	15
#define TLEN 100000
#define TNAME "pubkey.p15"



#if defined(TESTOUT) || defined(WIN32)
uint8_t RndSample[64] = {
	0x7b, 0xd5, 0x73, 0x33, 0x24, 0x48, 0xd4, 0x89, 
	0x76, 0xd8, 0x88, 0xb8, 0xf9, 0xde, 0x05, 0xab,  
	0x51, 0x32, 0x42, 0x8c, 0x3c, 0x12, 0x7c, 0xd1, 
	0x03, 0xf4, 0x6e, 0x76, 0x66, 0xc8, 0x74, 0x8d,
	0x9d, 0x61, 0x02, 0x27, 0xb9, 0xe0, 0x52, 0xc4, 
	0xeb, 0x0c, 0xb0, 0xcb, 0x13, 0x19, 0x57, 0xaf,
	0x41, 0x7f, 0x79, 0x2d, 0x7f, 0x20, 0xea, 0x4b, 
	0x34, 0xf7, 0xd1, 0x9c, 0xce, 0x14, 0x86, 0x90
};
#endif

/*****************************************************************************/
static void mydump(void* ptr, size_t len)
{
	size_t i, j;
	unsigned char c, *buf;
	unsigned int cc;

	j = 0;
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j && !(j & 3)) printf(" ");
		if (j > 15) {
			j = 0;
			printf("\n");
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		j++;
	}
	printf("\n");
}
/*****************************************************************************/
void print_errors(p15_info_t* errs)
{
	int i = 0;

	if (!errs) {
		printf("Success\n");
	}
	while (errs) {
		i++;
		printf("\n");
		printf("%d:\n", i);
		printf("Type: ");
		switch (errs->type) {
		case P15_ERROR:
			printf("Error\n");
			break;
		case P15_WARNING:
			printf("Warning\n");
			break;
		case P15_INTERNAL_ERROR:
			printf("Internal Error\n");
			break;
		default:
			printf("Unknown (%d)\n", errs->type);
		}
		printf("Code: ");
		if (errs->code < 0 || errs->code > P15_RC_MAX) 
			printf("Unknown (%d)\n", errs->code);
		else 
			printf("%s\n", p15_rc_dsc[errs->code]);
		if (errs->text) 
			printf("Details: %s\n", errs->text);
		if (errs->consumed)
			printf("Consumed bytes: %d\n", (int)errs->consumed);
		if (errs->asn_type_name)
			printf("ASN type name: %s\n", errs->asn_type_name);
		errs = errs->next;
	}
}
/*****************************************************************************/
int ckrc(char* func, int rc)
{
	if (!rc) return 0;

	printf("\n%s returned code: ", func);
	if (rc < 0 || rc > P15_RC_MAX) 
		printf("Unknown (%d)\n", rc);
	else 
		printf("%s\n", p15_rc_dsc[rc]);
	return rc;
}
/*****************************************************************************/
int ck_date(p15_utctime_t* in, int* gold, char* name)
{
	if (!in) {
		printf("%s date is NULL\n", name);
		return 1;
	}
	if (	gold[0] != in->year || 
		gold[1] != in->month ||
		gold[2] != in->day ||
		gold[3] != in->hours ||
		gold[4] != in->minutes ||
		gold[5] != in->seconds) {

		printf("%s date mismatches\n", name);
		return 1;
	}
	return 0;
}
/*****************************************************************************/
#if defined(TESTOUT) || defined(WIN32)
int init_fix_random(p15_ctx_t* ctx)
{
	return ckrc("p15_init_random", 
		p15_init_random(ctx, RndSample, RndSample + 32));
}
#endif
/*****************************************************************************/
int init_random(p15_ctx_t* ctx)
{
	int f = 0;
	uint8_t rnd[64];

	f = open("/dev/urandom", O_RDONLY);
	if (f < 0) {
		printf("Error opening /dev/urandom.\n");
		f = 0;
		return 1;
	}
	if (read(f, rnd, 64) != 64) {
		printf("Error reading /dev/urandom.\n");
		close(f);
		return 1;
	}
	close(f);
	f = 0;
	return ckrc("p15_init_random", p15_init_random(ctx, rnd, rnd + 32));
}
/*****************************************************************************/
void print_pubkey(p15_public_key_t* pubkey)
{
	printf("Public key:\n");
	printf("Label:'%s'\n", pubkey->label);
	if (pubkey->id.ext_id_len) {
		printf("ext_id:\n");
		mydump(pubkey->id.ext_id, pubkey->id.ext_id_len);
	} else {
		printf("id:0x%.8X\n", pubkey->id.id);
	}

	printf("Value:\n");
	mydump(pubkey->key, 64);

	printf("Key usage: 0x%.4X\n", pubkey->key_usage);

	printf("Key params: %d\n", pubkey->key_parameters);
	
	if (pubkey->start_date) {
		printf("Start date: %d-%d-%d %d:%.2d:%.2d\n", 
			pubkey->start_date->year,
			pubkey->start_date->month,
			pubkey->start_date->day,
			pubkey->start_date->hours,
			pubkey->start_date->minutes,
			pubkey->start_date->seconds);
	} else
		printf("Start date: none\n");

	if (pubkey->end_date) {
		printf("End date: %d-%d-%d %d:%.2d:%.2d\n", 
			pubkey->end_date->year,
			pubkey->end_date->month,
			pubkey->end_date->day,
			pubkey->end_date->hours,
			pubkey->end_date->minutes,
			pubkey->end_date->seconds);
	} else
		printf("End date: none\n");

	if (pubkey->subject) { 
		printf("Subject: present\n"); 
	} else
		printf("Subject: none\n");
}
/*****************************************************************************/
void print_privkey(p15_ctx_t* ctx, p15_private_key_t* prkey)
{
	int i;
	int rc;

	printf("Private key:\n");
	printf("Label:'%s'\n", prkey->label);
	if (prkey->id.ext_id_len) {
		printf("ext_id:\n");
		mydump(prkey->id.ext_id, prkey->id.ext_id_len);
	} else {
		printf("id:0x%.8X\n", prkey->id.id);
	}

	printf("Key masks: %d\n", (int)prkey->mask_count);
	for (i = 0; i < prkey->mask_count; i++) {
		printf("Key mask %d:\n", i);
		mydump(prkey->key_masks[i], 32);
	}

	rc = p15_remask_private_key(ctx, prkey, 1);
	if (rc) {
		printf("p15_remask_private_key: error %d\n", rc);
	}
	printf("Key unmasked:\n");
	mydump(prkey->key_masks[0], 32);

	printf("Key usage: 0x%.4X\n", prkey->key_usage);
	printf("Key access flags: 0x%.2X\n", prkey->key_access_flags);

	printf("Key params: %d\n", prkey->key_parameters);
	
	if (prkey->start_date) {
		printf("Start date: %d-%d-%d %d:%.2d:%.2d\n", 
			prkey->start_date->year,
			prkey->start_date->month,
			prkey->start_date->day,
			prkey->start_date->hours,
			prkey->start_date->minutes,
			prkey->start_date->seconds);
	} else
		printf("Start date: none\n");

	if (prkey->end_date) {
		printf("End date: %d-%d-%d %d:%.2d:%.2d\n", 
			prkey->end_date->year,
			prkey->end_date->month,
			prkey->end_date->day,
			prkey->end_date->hours,
			prkey->end_date->minutes,
			prkey->end_date->seconds);
	} else
		printf("End date: none\n");

	if (prkey->subject) { 
		printf("Subject: present\n"); 
	} else
		printf("Subject: none\n");

}
/*****************************************************************************/
int main(int argc, char** argv) {
	int rc = 0;
	p15_ctx_t* ctx = NULL;
	int f = 0;
	p15_info_t* errs = NULL;
	uint8_t token[TLEN];
	size_t tlen = TLEN;
	p15_utctime_t* sdate = NULL;
	p15_utctime_t* edate = NULL;
	p15_id_t* keyid = NULL;
	p15_object_t* obj = NULL;
	if (argc < 2) {
		printf("Usage: p15show token_file [passw]\n");
		return 1;
	}
	

	/* Loading */

	bzero(token, TLEN);
	f = open(argv[1], O_RDONLY|O_BINARY);
	if (f < 0) {
		f = 0;
		printf("Failed to open token.\n");
		rc = 1;
		goto end;
	};
	tlen = read(f, token, TLEN);
	close(f);
	f = 0;

	p15_free_ctx(&ctx);
	ctx = p15_new_ctx();

#if defined(TESTOUT) || defined(WIN32)
	rc = init_fix_random(ctx);
#else
	rc = init_random(ctx);
#endif

	if (rc) { rc = 1; goto end; }

	rc = ckrc("p15_load_token", p15_load_token(ctx, token, tlen,
		(uint8_t*)(argc == 3 ? argv[2] : NULL),
		argc == 3 ? strlen(argv[2]) : 0,
/*		NULL)); */
		&errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
/*	if (rc) { rc = 1; goto end; } */

/*
	if (!rc) {
		size_t tlen_tmp = 0;
		printf("Renewing the token (not storing)...\n");
		rc = ckrc("p15_save_token", p15_save_token(ctx, 
			token, &tlen_tmp, 
			(uint8_t*)(argc == 3 ? argv[2] : NULL),
			argc == 3 ? strlen(argv[2]) : 0,
			&errs));
		if (errs) {
			print_errors(errs);
			p15_free_infos(&errs);
		}
		if (rc > 1) { rc = 1; goto end; }
		rc = 0;
	}
*/
	rc = 0;
	
	obj = ctx->objects;
	while (obj) {
		switch (obj->type) {
		case P15_OT_PRIVKEY:
			print_privkey(ctx, (p15_private_key_t*)obj->obj);
			break;

		case P15_OT_PUBKEY:
			print_pubkey((p15_public_key_t*)obj->obj);
			break;

		default:
			printf("Object type %d not implemented yet.\n", obj->type);
		}
		obj = obj->next;
		printf("\n");
	}
end:
	if (f) close(f);
	p15_free_time(&sdate);
	p15_free_time(&edate);
	if (errs) 
		print_errors(errs);
	p15_free_infos(&errs);
	p15_free_ctx(&ctx);
	p15_free_id(&keyid);
/*	if (rc) 
		printf("TEST FAILED\n");
	else
		printf("TEST PASSED\n");
*/
	return rc;
}

