


#if defined(WIN32) || defined(WIN64) /* Windows */

#include "stdafx.h"
#include <io.h>
#include <windows.h>
#include "pkcs15dll.h"

#ifndef bzero
#define bzero ZeroMemory
#endif 

#define p15t_open  _open
#define p15t_read  _read
#define p15t_write _write
#define p15t_close _close

#else /* Linux */

#include <strings.h>
#include <unistd.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define p15t_open  open
#define p15t_read  read
#define p15t_write write
#define p15t_close close

#endif /* defined(WIN32) || defined(WIN64) */


#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <sys/types.h>

#include <pkcs15container.h>

#define RND_EXTRA_LEN	15
#define TLEN 100000
#define TNAME_2001 "token2001.p15"
#define TNAME_2012 "token2012.p15"

/*****************************************************************************/
char		CACertFile[] = "ca.cer";
char		CACertLabel[] = "Root Certificate of CA";
char		CRLFile[] = "crl";
char		CRLLabel[] = "CRL from CA";
char		FedotovCertFile[] = "fedotov.cer";
char		FedotovCertLabel[] = "Certificate of Andrey Fedotov";
char		FedotovKeyFile[] = "fedotov.key";
char		FedotovKeyLabel[] = "Private Key of Andrey Fedotov";
int		FedotovKeyParams = P15_PKIPAR_CPRO_A;
int		FedotovKeyStartDate[6] = { 2012, 5, 18, 11, 3, 0 };
int 		FedotovKeyEndDate[6] = { 2013, 5, 18, 11, 12, 0 };
uint16_t 	FedotovKeyUsage = P15_KU_SIGN | P15_KU_NON_REPUDIATION | P15_KU_UNWRAP
			| P15_KU_DECRYPT;

char		PrKey2File[] = "prkey2";
char		PrKey2Label[] = "New Generated Private Key";
int		Key2Params = P15_PKIPAR_CPRO_B;
int		PrKey2Usage = P15_KU_SIGN;
char		PubKey2File[] = "pubkey2";
char		PubKey2Label[] = "Public Key for New Generated Private Key";
int		PubKey2Usage = P15_KU_VERIFY;

char		PrKey64File[] = "prkey64";
char		PrKey64Label[] = "512-bit private key";
int		Key64Params = P15_PKIPAR_TC26_512_TEST;
int		PrKey64Usage = P15_KU_SIGN;
char		PubKey64File[] = "pubkey64";
char		PubKey64Label[] = "Public key for 512-bit private key";
int		PubKey64Usage = P15_KU_VERIFY;
int		Key64StartDate[6] = { 2013, 10, 22, 12, 14, 33 };
int		Key64EndDate[6] = { 2014, 10, 18, 22, 0, 0 };

int 		MaskCount = 3;


char		SecKey1File[] = "seckey1";
char		SecKey1Label[] = "Gost89 key, CproB";
int		SecKey1Params = P15_89UZ_CPRO_Z;
int		SecKey1Usage = P15_KU_ENCRYPT | P15_KU_DECRYPT;
int		SecKey1StartDate[6] = { 2014, 4, 17, 11, 50 };
int		SecKey1EndDate[6] = { 2015, 5, 31, 23, 59, 59 };
int		SecKey1MaskCount = 3;

char		SecKey2File[] = "seckey2";
char		SecKey2Label[] = "Sym Gost89 key 2, CproA";
int		SecKey2Params = P15_89UZ_CPRO_C;
int		SecKey2Usage = P15_KU_SIGN | P15_KU_VERIFY;
int		SecKey2MaskCount = 10;

int		EncUZ[2] = { P15_89UZ_CPRO_C, P15_89UZ_CPRO_Z }; /* P15-2001, P15-2012 */


char	JustString1[] = 
"This is some confidential abstract data. It will be encrypted in the container.";

char	JustString2[] =
"This is some open data. There's no need to encrypt it.";


#define PRKEY_AFLAGS	P15_KA_DEFAULT

#if defined(TESTOUT) || defined(WIN32)
uint8_t RndSample[64] = {
	0x7b, 0xd5, 0x73, 0x33, 0x24, 0x48, 0xd4, 0x89, 
	0x76, 0xd8, 0x88, 0xb8, 0xf9, 0xde, 0x05, 0xab,  
	0x51, 0x32, 0x42, 0x8c, 0x3c, 0x12, 0x7c, 0xd1, 
	0x03, 0xf4, 0x6e, 0x76, 0x66, 0xc8, 0x74, 0x8d,
	0x9d, 0x61, 0x02, 0x27, 0xb9, 0xe0, 0x52, 0xc4, 
	0xeb, 0x0c, 0xb0, 0xcb, 0x13, 0x19, 0x57, 0xaf,
	0x41, 0x7f, 0x79, 0x2d, 0x7f, 0x20, 0xea, 0x4b, 
	0x34, 0xf7, 0xd1, 0x9c, 0xce, 0x14, 0x86, 0x90
};
#endif

char OidSt1[] = "1.1.456.7890";
char OidSt2[] = "1.1.765.432.1";

/*****************************************************************************/
void print_errors(p15_info_t* errs)
{
	int i = 0;

	if (!errs) {
		printf("Success\n");
	}
	while (errs) {
		i++;
		printf("\n");
		printf("%d:\n", i);
		printf("Type: ");
		switch (errs->type) {
		case P15_ERROR:
			printf("Error\n");
			break;
		case P15_WARNING:
			printf("Warning\n");
			break;
		case P15_INTERNAL_ERROR:
			printf("Internal Error\n");
			break;
		default:
			printf("Unknown (%d)\n", errs->type);
		}
		printf("Code: ");
		if (errs->code < 0 || errs->code > P15_RC_MAX) 
			printf("Unknown (%d)\n", errs->code);
		else 
			printf("%s\n", p15_rc_dsc[errs->code]);
		if (errs->text) 
			printf("Details: %s\n", errs->text);
		if (errs->consumed)
			printf("Consumed bytes: %d\n", (int)errs->consumed);
		if (errs->asn_type_name)
			printf("ASN type name: %s\n", errs->asn_type_name);
		errs = errs->next;
	}
}
/*****************************************************************************/
int ckrc(char* func, int rc)
{
	if (!rc) return 0;

	printf("\n%s returned code: ", func);
	if (rc < 0 || rc > P15_RC_MAX) 
		printf("Unknown (%d)\n", rc);
	else 
		printf("%s\n", p15_rc_dsc[rc]);
	return rc;
}
/*****************************************************************************/
int ck_date(p15_utctime_t* in, int* gold, char* name)
{
	if ((!in && gold) || (in && !gold)) {
		printf("%s date mismatches (absent/present)\n", name);
		return 1;
	}
	if (!in) return 0;
	if (	gold[0] != in->year || 
		gold[1] != in->month ||
		gold[2] != in->day ||
		gold[3] != in->hours ||
		gold[4] != in->minutes ||
		gold[5] != in->seconds) {

		printf("%s date mismatches\n", name);
		return 1;
	}
	return 0;
}
/*****************************************************************************/
#if defined(TESTOUT) || defined(WIN32) || defined(WIN64)
int init_fix_random(p15_ctx_t* ctx)
{
	return ckrc("p15_init_random", 
		p15_init_random(ctx, RndSample, RndSample + 32));
}

#else /***********************************************************************/

int init_random(p15_ctx_t* ctx)
{
	int f = -1;
	uint8_t rnd[64];
	f = open("/dev/urandom", O_RDONLY);
	if (f < 0) {
		printf("Error opening /dev/urandom.\n");
		return 1;
	}
	if (read(f, rnd, 64) != 64) {
		printf("Error reading /dev/urandom.\n");
		close(f);
		return 1;
	}
	close(f);
	return ckrc("p15_init_random", p15_init_random(ctx, rnd, rnd + 32));
}
#endif
/*****************************************************************************/
int add_private_key(p15_ctx_t* ctx, char* keyfile, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, int key_params, int* start_date, int* end_date,
	int mask_count)
{
	int f = -1;
	uint8_t kbuf[64];
	int rc = 1;
	int rc2;
	p15_info_t* errs = NULL;
	p15_utctime_t* stime = NULL;
	p15_utctime_t* etime = NULL;
	int *sd = start_date;
	int *ed = end_date;
	p15_private_key_t* key = NULL;
	p15_private_key_t* key1 = NULL;
	p15_object_t* it = NULL;
	int is2012;

	is2012 = key_params >= P15_PKIPAR_TC26_512_TEST;

	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, kbuf, is2012 ? 64 : 32) != (is2012 ? 64 : 32)) {
		printf("Key file %s is not %s bytes.\n", keyfile, is2012 ? "64" : "32");
		goto end;
	}

	if (sd) stime = p15_new_time(sd[0], sd[1], sd[2], sd[3], sd[4], sd[5]);
	if (ed) etime = p15_new_time(ed[0], ed[1], ed[2], ed[3], ed[4], ed[5]);

	rc2 = ckrc("p15_add_private_key", p15_add_private_key(ctx, id, newid, label,
		key_usage, PRKEY_AFLAGS, key_params, stime, etime, &errs, 1, kbuf));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc2) goto end;
	
	key1 = p15_get_next_private_key(ctx, &it);
	while (key1) {
		key = key1;
		key1 = p15_get_next_private_key(ctx, &it);
	}
	if (!key) {
		printf("Error: add_private_key: Internal error 1.\n");
		goto end;
	}

	rc2 = ckrc("p15_remask_setup", p15_remask_setup(ctx, key, NULL, 0, mask_count, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc2) goto end;

	rc = 0;
end:
	if (f > -1) 
		p15t_close(f);
	p15_free_time(&stime);
	p15_free_time(&etime);
	p15_free_infos(&errs);
	return rc;
}
/*****************************************************************************/
int ck_private_key(p15_ctx_t* ctx, char* keyfile, p15_private_key_t* pkey,
	char* label, uint16_t key_usage, int key_params, int* start_date, int* end_date,
	int mask_count, int subj_exists)
{
	int rc = 1;
	int rc2;
	uint8_t kbuf[64];
	int f = -1;
	int is2012;

	is2012 = key_params >= P15_PKIPAR_TC26_512_TEST;

	if (!pkey) {
		printf("Error: Key object expected\n");
		goto end;
	}

	printf("Private key:'%s'\n", pkey->label);
	if (strcmp(label, pkey->label)) {
		printf("Unexpected label. Expected:%s\n", label);
		goto end;
	}
	printf("id:%d\n", pkey->id.id);
	if (pkey->key_usage != key_usage) {
		printf("Wrong key usage\n");
		goto end;
	}
	if (pkey->key_parameters != key_params) {
		printf("Wrong key params\n");
		goto end;
	}

	if (pkey->key_access_flags != PRKEY_AFLAGS) {
		printf("Wrong key access flags\n");
		goto end;
	}

	if (ck_date(pkey->start_date, start_date, "start")) {
		 goto end;
	}
	if (ck_date(pkey->end_date, end_date, "end")) {
		goto end;
	}

	if (pkey->mask_count != mask_count) {
		printf("Wrong mask number\n");
		goto end;
	}

	if ((subj_exists && !pkey->subject) || (!subj_exists && pkey->subject)) {
		printf("Subject presence/absence mismatch\n");
		goto end;
	}

	rc2 = ckrc("p15_remask_private_key", p15_remask_private_key(ctx, pkey, 1));
	if (rc2) goto end;

	if (pkey->mask_count != 1) {
		printf("Wrong mask number (after unmasking)\n");
		goto end;
	}

	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, kbuf, is2012 ? 64 : 32) != (is2012 ? 64 : 32)) {
		printf("Key file %s is not %s bytes.\n", keyfile, is2012 ? "64" : "32");
		goto end;
	}
	if (memcmp(pkey->key_masks[0], kbuf, is2012 ? 64 : 32)) {
		printf("Private key mismatches\n");
		goto end;
	}

	rc = 0;

end:
	if (f > -1) p15t_close(f);
	return rc;
}
/*****************************************************************************/
int add_secret_key(p15_ctx_t* ctx, char* keyfile, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, int key_params, int* start_date, int* end_date,
	int mask_count)
{
	int f = -1;
	uint8_t kbuf[32];
	int rc = 1;
	int rc2;
	p15_info_t* errs = NULL;
	p15_utctime_t* stime = NULL;
	p15_utctime_t* etime = NULL;
	int *sd = start_date;
	int *ed = end_date;
	p15_secret_key_t* key = NULL;
	p15_secret_key_t* key1 = NULL;
	p15_object_t* it = NULL;

	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, kbuf, 32) != 32) {
		printf("Key file %s is not %s bytes.\n", keyfile, "32");
		goto end;
	}
	if (sd) stime = p15_new_time(sd[0], sd[1], sd[2], sd[3], sd[4], sd[5]);
	if (ed) etime = p15_new_time(ed[0], ed[1], ed[2], ed[3], ed[4], ed[5]);

	rc2 = ckrc("p15_add_secret_key", p15_add_secret_key(ctx, id, newid, label,
		key_usage, PRKEY_AFLAGS, key_params, stime, etime, &errs, 1, kbuf));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc2) goto end;
	
	key1 = p15_get_next_secret_key(ctx, &it);
	while (key1) {
		key = key1;
		key1 = p15_get_next_secret_key(ctx, &it);
	}
	if (!key) {
		printf("Error: add_secret_key: Internal error 1.\n");
		goto end;
	}

	rc2 = ckrc("p15_remask_setup", p15_remask_setup(ctx, NULL, key, 0, mask_count, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc2) goto end;

	rc = 0;
end:
	if (f > -1) p15t_close(f);
	p15_free_time(&stime);
	p15_free_time(&etime);
	p15_free_infos(&errs);
	return rc;
}

/*****************************************************************************/
int ck_secret_key(p15_ctx_t* ctx, char* keyfile, p15_secret_key_t* skey,
	char* label, uint16_t key_usage, int key_params, int* start_date, int* end_date,
	int mask_count)
{
	int rc = 1;
	int rc2;
	uint8_t kbuf[32];
	int f = -1;

	if (!skey) {
		printf("Error: Secret key object expected\n");
		goto end;
	}

	printf("Secret key:'%s'\n", skey->label);
	if (strcmp(label, skey->label)) {
		printf("Unexpected label. Expected:%s\n", label);
		goto end;
	}
	printf("id:%d\n", skey->id.id);
	if (skey->key_usage != key_usage) {
		printf("Wrong key usage\n");
		goto end;
	}
	if (skey->key_parameters != key_params) {
		printf("Wrong key params\n");
		goto end;
	}

	if (skey->key_access_flags != PRKEY_AFLAGS) {
		printf("Wrong key access flags\n");
		goto end;
	}

	if (ck_date(skey->start_date, start_date, "start")) {
		 goto end;
	}
	if (ck_date(skey->end_date, end_date, "end")) {
		goto end;
	}
	if (skey->mask_count != mask_count) {
		printf("Wrong mask number\n");
		goto end;
	}

	rc2 = ckrc("p15_remask_secret_key", p15_remask_secret_key(ctx, skey, 1));
	if (rc2) goto end;

	if (skey->mask_count != 1) {
		printf("Wrong mask number (after unmasking)\n");
		goto end;
	}
	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, kbuf, 32) != 32) {
		printf("Key file %s is not %s bytes.\n", keyfile, "32");
		goto end;
	}
	if (memcmp(skey->key_masks[0], kbuf, 32)) {
		printf("Secret key mismatches\n");
		goto end;
	}

	rc = 0;
end:
	if (f > -1) p15t_close(f);
	return rc;
}
/*****************************************************************************/
int add_public_key(p15_ctx_t* ctx, char* keyfile, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, int key_params,
	int* start_date, int* end_date)
{
	int f = -1;
	uint8_t key[128];
	int rc = 1;
	int rc2;
	p15_info_t* errs = NULL;
	p15_utctime_t* stime = NULL;
	p15_utctime_t* etime = NULL;
	int *sd = start_date;
	int *ed = end_date;
	int is2012;

	is2012 = key_params >= P15_PKIPAR_TC26_512_TEST;

	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, key, is2012 ? 128 : 64) != (is2012 ? 128 :64)) {
		printf("Key file %s is not %s bytes.\n", keyfile, is2012 ? "128" : "64");
		goto end;
	}
	if (sd) stime = p15_new_time(sd[0], sd[1], sd[2], sd[3], sd[4], sd[5]);
	if (ed) etime = p15_new_time(ed[0], ed[1], ed[2], ed[3], ed[4], ed[5]);

	rc2 = ckrc("p15_add_public_key", p15_add_public_key(ctx, id, newid, label, 
		key, key_usage, key_params, stime, etime, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc2) goto end;

	rc = 0;
end:
	if (f > -1) p15t_close(f);
	p15_free_time(&stime);
	p15_free_time(&etime);
	p15_free_infos(&errs);
	return rc;
}
/*****************************************************************************/
int ck_public_key(char* keyfile, p15_public_key_t* pubkey, char* label, 
	uint16_t key_usage, int key_params, int* start_date, int* end_date)
{
	int rc = 1;
	uint8_t kbuf[128];
	int f = -1;
	int is2012;

	if (!pubkey) {
		printf("Public key object expected\n");
		goto end;
	}

	is2012 = key_params >= P15_PKIPAR_TC26_512_TEST;

	printf("Public key loaded:'%s'\n", pubkey->label);
	if (strcmp(label, pubkey->label)) {
		printf("Unexpected label. Expected:%s\n", label);
		goto end;
	}

	printf("id:%d\n", pubkey->id.id);
	if (pubkey->key_usage != key_usage) {
		printf("Wrong key usage\n");
		goto end;
	}
	if (pubkey->key_parameters != key_params) {
		printf("Wrong key params\n");
		goto end;
	}
	if (ck_date(pubkey->start_date, start_date, "start")) {
		goto end;
	}
	if (ck_date(pubkey->end_date, end_date, "end")) {
		goto end;
	}

	f = p15t_open(keyfile, O_RDONLY | O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", keyfile);
		goto end;
	}
	if (p15t_read(f, kbuf, is2012 ? 128 : 64) != (is2012 ? 128 : 64)) {
		printf("Key file %s is not %s bytes.\n", keyfile, is2012 ? "128" : "64");
		goto end;
	}
	if (memcmp(pubkey->key, kbuf, is2012 ? 128 : 64)) {
		printf("Public key mismatches\n");
		goto end;
	}
	rc = 0;
end:
	if (f > -1) p15t_close(f);
	return rc;
}
/*****************************************************************************/
int add_cert(p15_ctx_t* ctx, char* certfile, int trusted, p15_id_t* id, p15_id_t** newid, 
	char* label)
{
	int f = -1;
	int l;
	uint8_t buf[TLEN];
	int rc;
	p15_info_t* errs = NULL;
	f = p15t_open(certfile, O_RDONLY|O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", certfile);
		return 1;
	}
	l = p15t_read(f, buf, TLEN);
	p15t_close(f);
	rc = ckrc("p15_add_cert", p15_add_cert(ctx, trusted, trusted, id, newid,
		label, buf, l, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc) return 1;
	
	return 0;			
}
/*****************************************************************************/
int ck_cert(p15_certificate_t* cert, char* certfile)
{
	int f = -1;
	uint8_t buf[TLEN];
	int l;

	if (!cert) {
		printf("Certificate object expected.\n");
		return 1;
	}

	printf("\nCertificate: '%s'\n", cert->label);
	printf("id: %d\n", cert->id.id);
	printf("ca: %d\n", cert->ca);
	if (!cert->subject || !cert->issuer) {
		printf("subject of issuer is NULL\n");
		return 1;
	}
	f = p15t_open(certfile, O_RDONLY|O_BINARY);
	if (f < 0) { 
		printf("%s file not found\n", certfile);
		return 1;
	}
	l = p15t_read(f, buf, TLEN);
	p15t_close(f);
	if (l != cert->len) {
		printf("cert size mismatch\n");
		return 1;
	}
	if (memcmp(cert->ptr, buf, l)) {
		printf("cert data mismatch\n");
		return 1;
	}
	return 0;
}
/*****************************************************************************/
int add_crl(p15_ctx_t* ctx, char* crlfile, char* label, p15_id_t* id)
{
	int f = -1;
	int l;
	uint8_t buf[TLEN];
	int rc;
	p15_info_t* errs = NULL;
	f = p15t_open(crlfile, O_RDONLY|O_BINARY);
	if (f < 0) {
		printf("Cannot read file %s\n", crlfile);
		return 1;
	}
	l = p15t_read(f, buf, TLEN);
	p15t_close(f);
	rc = ckrc("p15_add_crl", p15_add_crl(ctx, id, NULL, label, buf, l, 0,
		&errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	if (rc) return 1;
	
	return 0;	
}
/*****************************************************************************/
int ck_crl(p15_crl_t* crl, char* crlfile)
{
	int f = -1;
	uint8_t buf[TLEN];
	int l;

	if (!crl) {
		printf("CRL object expected.\n");
		return 1;
	}
	printf("\nCRL: '%s'\n", crl->label);
	printf("id: %d\n", crl->id.id);
	if (!crl->issuer) {
		printf("issuer is NULL\n");
		return 1;
	}
	f = open(crlfile, O_RDONLY|O_BINARY);
	if (f < 0) { 
		printf("%s file not found\n", crlfile);
		return 1;
	}
	l = read(f, buf, TLEN);
	close(f);
	if (l != crl->len) {
		printf("crl size mismatch\n");
		return 1;
	}
	if (memcmp(crl->ptr, buf, l)) {
		printf("crl data mismatch\n");
		return 1;
	}
	return 0;
}
/*****************************************************************************/
void* get_next_obj(p15_ctx_t* ctx, p15_object_t** obj, int type)
{
	p15_object_t* o;

	if (!obj || !ctx) {
		printf("get_next_obj: error 1\n");
		return NULL;
	}
	o = *obj;
	if (!o) o = ctx->objects;
	if (!o) { *obj = NULL; return NULL; }
	*obj = o->next;
	if (type != o->type) {
		printf ("get_next_obj: Object type mismatch (%d, must be %d)\n",
			o->type, type);
		return NULL;
	}
	return o->obj;
}
/*****************************************************************************/
#ifndef TESTOUT
static int ck_oiddo(p15_oiddo_t* oiddo, char* oidst, uint8_t* data,
	int len, int to_encrypt)
{
	p15_oid_t* oid = NULL;
	int rc = 0;

	oid = p15_new_oid_from_str(oidst);
	printf("Label:'%s'\n", oiddo->label);
	rc = p15_oid_cmp(oid, oiddo->oid);
	if (rc) {
		printf("OIDs mismatch\n");
		goto end;
	}
	if (oiddo->len != len) {
		printf("Data lengths mismatch\n");
		rc = 1;
		goto end;
	}
	if (memcmp(oiddo->ptr, data, len)) {
		printf("Data mismatch\n");
		rc = 1;
		goto end;
	}
	if (to_encrypt != oiddo->to_encrypt) {
		printf("Encrypt flag mismatch\n");
		rc = 1;
		goto end;
	}
end:
	p15_free_oid(&oid);
	return rc;
}
#endif
/*****************************************************************************/
int DoTest(int rd, int is2012, char* passw)
{
	int rc = 1;
	p15_ctx_t* ctx = NULL;
	int f = -1;
	p15_info_t* errs = NULL;
	uint8_t token[TLEN];
	size_t tlen = TLEN;
	p15_id_t* id = NULL;
	p15_oid_t* oidt = NULL;
#ifndef TESTOUT
	p15_private_key_t* pkey = NULL;
	p15_public_key_t* pubkey = NULL;
	p15_certificate_t* cert = NULL;
	p15_crl_t* crl = NULL;
	p15_object_t* obj = NULL;
	p15_oiddo_t* oiddo = NULL;
	p15_secret_key_t* skey = NULL;
#endif

	if (!rd) {
		ctx = p15_new_ctx();
		if (!ctx) { 
			rc = 1;
			printf("Error creating ctx.\n");
			goto end;
		}
#if defined(TESTOUT) || defined(WIN32) || defined(WIN64)
		rc = init_fix_random(ctx);
#else
		rc = init_random(ctx);
#endif
		/* Extra random len */
		
		if (!rc) rc = ckrc("p15_random_setup",
			p15_random_setup(ctx, NULL, RND_EXTRA_LEN + 64));
		if (rc) goto end;

		/* EncUZ */
		if (!rc) rc = ckrc("p15_set_default_enc_uz",
			p15_set_default_enc_uz(ctx, EncUZ[!!is2012]));
		if (rc) goto end;

		/* CA cert and CRL */
		if (!rc) rc = add_cert(ctx, CACertFile, 1, NULL, &id, CACertLabel);
		if (!rc) rc = add_crl(ctx, CRLFile, CRLLabel, id);
		if (rc) goto end;
		p15_free_id(&id);

		/* Fedotov cert and key */
		rc = add_private_key(ctx, FedotovKeyFile, NULL, &id, FedotovKeyLabel, 
			FedotovKeyUsage, FedotovKeyParams,
			FedotovKeyStartDate, FedotovKeyEndDate, MaskCount);
		if (!rc) rc = add_cert(ctx, FedotovCertFile, 0, id, NULL, FedotovCertLabel);
		if (!rc) rc = ckrc("p15_set_private_key_subject", 
			p15_set_pkey_subject(ctx, id, NULL));
		if (rc) goto end;
		p15_free_id(&id);

		/* New pub and priv key */
		rc = add_private_key(ctx, PrKey2File, NULL, &id, PrKey2Label,
			PrKey2Usage, Key2Params,
			NULL, NULL, MaskCount);
		if (!rc) rc = add_public_key(ctx, PubKey2File, id, NULL, PubKey2Label,
			PubKey2Usage, Key2Params,
			NULL, NULL);
		if (rc) goto end;
		p15_free_id(&id);

		if (is2012) { /* 2012 priv and pub key */
			rc = add_private_key(ctx, PrKey64File, NULL, &id, PrKey64Label,
				PrKey64Usage, Key64Params,
				Key64StartDate, Key64EndDate, MaskCount);
			if (!rc) rc = add_public_key(ctx, PubKey64File, id, NULL, 
				PubKey64Label, PubKey64Usage, Key64Params,
				Key64StartDate, Key64EndDate);
			if (rc) goto end;
			p15_free_id(&id);
		}

		/* oiddo */
		rc = ckrc("p15_add_oiddo", 
			p15_add_oiddo(ctx, "Top-secret Data", OidSt1, NULL,
			(uint8_t*)JustString1, sizeof(JustString1), 1, &errs));
		if (rc) goto end;

		oidt = p15_new_oid_from_str(OidSt2);
		rc = ckrc("p15_add_oiddo", 
			p15_add_oiddo(ctx, "Public Data", NULL, oidt,
			(uint8_t*)JustString2, sizeof(JustString2), 0, &errs));
		if (rc) goto end;
	
		if (is2012) {
			rc = add_secret_key(ctx, SecKey1File, NULL, NULL, SecKey1Label,
				SecKey1Usage, SecKey1Params, 
				SecKey1StartDate, SecKey1EndDate, SecKey1MaskCount);
			if (rc) goto end;
#ifndef TESTOUT
			id = p15_new_id(100);
			if (!id) {
				printf("DoTest: p15_new_id error 1\n");
				rc = 1;
				goto end;
			}
			rc = add_secret_key(ctx, SecKey2File, id, NULL, SecKey2Label,
				SecKey2Usage, SecKey2Params,
				NULL, NULL, SecKey2MaskCount);
			p15_free_id(&id);
			if (rc) goto end;
#endif
		}

		if (!is2012) {
			rc = ckrc("p15_set_format", p15_set_format(ctx, P15_FORMAT_2001));
			if (rc) goto end;
		}

		rc = ckrc("p15_save_token", p15_save_token(ctx, 
			token, &tlen, (uint8_t*)passw, passw ? strlen(passw) : 0, &errs));
		if (errs) {
			print_errors(errs);
			p15_free_infos(&errs);
		}
		if (rc > 1) { rc = 1; goto end; }
		rc = 0;
	
		unlink(is2012 ? TNAME_2012 : TNAME_2001);
		f = p15t_open(is2012 ? TNAME_2012 : TNAME_2001, O_WRONLY | O_CREAT|O_BINARY, 0644);
		if (f < 0) {
			printf("Failed to create token file.\n");
			rc = 1;
			goto end;
		}
		if (p15t_write(f, token, tlen) != tlen) {
			printf("Failed to write token.\n");
			rc = 1;
			goto end;
		}
		p15t_close(f);
	} /* !rd */

#ifndef TESTOUT
	/* Loading */

	bzero(token, TLEN);

	f = p15t_open(is2012 ? TNAME_2012 : TNAME_2001, O_RDONLY|O_BINARY);
	if (f < 0) {
		printf("Failed to open token.\n");
		rc = 1;
		goto end;
	};
	tlen = p15t_read(f, token, TLEN);
	p15t_close(f);
	f = -1;/*##Ivan: In Visual Studio after _close f IS NOT <0. That's why an error was later!*/

	p15_free_ctx(&ctx);
	ctx = p15_new_ctx();

#if defined(TESTOUT) || defined(WIN32) || defined(WIN64)
	rc = init_fix_random(ctx);
#else
	rc = init_random(ctx);
#endif

	if (rc) { rc = 1; goto end; }

	rc = ckrc("p15_load_token", p15_load_token(ctx, token, tlen,
		(uint8_t*)passw, passw ? strlen(passw) : 0, &errs));
	if (errs) {
		print_errors(errs);
		p15_free_infos(&errs);
	}
	
	if (is2012) {
		if (ctx->p15_format != P15_FORMAT_2012) {
			printf("Error: P15-2012 format expected.\n");
			rc = 1;
			goto end;
		}
	} else {
		if (ctx->p15_format != P15_FORMAT_2001) {
			printf("Error: P15-2001 format expected.\n");
			rc = 1;
			goto end;
		}
	}

	if (ctx->enc_uz != EncUZ[!!is2012]) {
		printf("Error: ctx->enc_uz mismatch.\n");
		rc = 1;
		goto end;
	}

	printf("\nLoaded token %d.%d\n\n", ctx->version_major, ctx->version_minor);

	if (!rc) {
		size_t tlen_tmp = 0;
		printf("Renewing the token (not storing)...\n");
		rc = ckrc("p15_save_token", p15_save_token(ctx, 
			token, &tlen_tmp, (uint8_t*)passw, passw ? strlen(passw) : 0, &errs));
		if (errs) {
			print_errors(errs);
			p15_free_infos(&errs);
		}
		if (rc > 1) { rc = 1; goto end; }
		rc = 0;
	}
	rc = 0;

	printf("\n");


	cert = (p15_certificate_t *)get_next_obj(ctx, &obj, P15_OT_TRUSTED_CERT);
	rc = ck_cert(cert, CACertFile);
	if (rc) goto end;
	printf("\n");

	crl = (p15_crl_t *)get_next_obj(ctx, &obj, P15_OT_CRL);
	rc = ck_crl(crl, CRLFile);	
	if (rc) goto end;
	printf("\n");

	pkey = (p15_private_key_t*)get_next_obj(ctx, &obj, P15_OT_PRIVKEY);
	rc = ck_private_key(ctx, FedotovKeyFile, pkey, FedotovKeyLabel, FedotovKeyUsage,
		FedotovKeyParams, FedotovKeyStartDate, FedotovKeyEndDate, MaskCount, 1);
	if (rc) goto end;
	printf("\n");

	cert = (p15_certificate_t *)get_next_obj(ctx, &obj, P15_OT_CERT);
	rc = ck_cert(cert, FedotovCertFile);
	if (rc) goto end;
	printf("\n");
	
	pkey = (p15_private_key_t*)get_next_obj(ctx, &obj, P15_OT_PRIVKEY);
	rc = ck_private_key(ctx, PrKey2File, pkey, PrKey2Label, PrKey2Usage,
		Key2Params, NULL, NULL, MaskCount, 0);
	if (rc) goto end;
	printf("\n");
	
	pubkey = (p15_public_key_t *)get_next_obj(ctx, &obj, P15_OT_PUBKEY);
	rc = ck_public_key(PubKey2File, pubkey, PubKey2Label, PubKey2Usage,
		Key2Params, NULL, NULL);
	if (rc) goto end;
	printf("\n");

	if (is2012) {
		pkey = (p15_private_key_t*)get_next_obj(ctx, &obj, P15_OT_PRIVKEY);
		rc = ck_private_key(ctx, PrKey64File, pkey, PrKey64Label, PrKey64Usage,
			Key64Params, Key64StartDate, Key64EndDate, MaskCount, 0);
		if (rc) goto end;
		printf("\n");
	
		pubkey = (p15_public_key_t *)get_next_obj(ctx, &obj, P15_OT_PUBKEY);
		rc = ck_public_key(PubKey64File, pubkey, PubKey64Label, PubKey64Usage,
			Key64Params, Key64StartDate, Key64EndDate);
		if (rc) goto end;
		printf("\n");
	}

	oiddo = (p15_oiddo_t*)get_next_obj(ctx, &obj, P15_OT_OIDDO);
	rc = ck_oiddo(oiddo, OidSt1, (uint8_t*)JustString1, sizeof(JustString1), 1);
	if (rc) goto end;
	printf("\n");

	oiddo = (p15_oiddo_t*)get_next_obj(ctx, &obj, P15_OT_OIDDO);
	rc = ck_oiddo(oiddo, OidSt2, (uint8_t*)JustString2, sizeof(JustString2), 0);
	if (rc) goto end;
	printf("\n");

	if (is2012) {
		skey = (p15_secret_key_t*)get_next_obj(ctx, &obj, P15_OT_SECKEY);
		rc = ck_secret_key(ctx, SecKey1File, skey, SecKey1Label, SecKey1Usage,
			SecKey1Params, SecKey1StartDate, SecKey1EndDate, SecKey1MaskCount);
		if (rc) goto end;
		printf("\n");

		skey = (p15_secret_key_t*)get_next_obj(ctx, &obj, P15_OT_SECKEY);
		rc = ck_secret_key(ctx, SecKey2File, skey, SecKey2Label, SecKey2Usage,
			SecKey2Params, NULL, NULL, SecKey2MaskCount);
		if (rc) goto end;
		printf("\n");
	}

	if (obj) {
		printf("Error: some other objects left in the list.\n");
		rc = 1;
		goto end;
	}
#endif /* ifndef TESTOUT */

	rc = 0;
end:
	if (f > -1) p15t_close(f);
	if (errs) 
		print_errors(errs);
	p15_free_infos(&errs);
	p15_free_ctx(&ctx);
	p15_free_id(&id);
	p15_free_oid(&oidt);

	return rc;
}
/*****************************************************************************/
int main(int argc, char** argv) {
	int rc = 0;
	int rd = 0;

	if (argc < 2) {
		printf("Usage: test1 r|w [passw]\n");
		return 1;
	}
	
	if (argv[1][0] == 'r') rd = 1;

#ifndef TESTOUT
	rc = DoTest(rd, 0, argc < 3 ? NULL : argv[2]);
	if (rc) goto end;
#endif
	rc = DoTest(rd, 1, argc < 3 ? NULL : argv[2]);
	if (rc) goto end;

end:
	if (rc) 
		printf("TEST FAILED\n");
	else
		printf("TEST PASSED\n");
#if defined(WIN32) || defined(WIN64)
	system("PAUSE");/*##Ivan: Press Any Key before close this console-window. */
#endif
	return rc;
}

